"""Validation utilities for cnsplots."""

from typing import Any, List, Optional, Union

import numpy as np
import pandas as pd

# ============================================================================
# Input Type Validation
# ============================================================================


def validate_dataframe(data: Any, param_name: str, function_name: str) -> None:
    """
    Validate that data is a pandas DataFrame.

    Parameters
    ----------
    data : Any
        The data to validate.
    param_name : str
        The name of the parameter being validated.
    function_name : str
        The name of the calling function for error messages.

    Raises
    ------
    TypeError
        If data is not a pandas DataFrame.
    """
    if not isinstance(data, pd.DataFrame):
        raise TypeError(
            f"[{function_name}] Parameter '{param_name}' must be a pandas DataFrame, "
            f"got {type(data).__name__}"
        )


def validate_anndata(adata: Any, param_name: str, function_name: str) -> None:
    """
    Validate that adata is an AnnData object.

    Parameters
    ----------
    adata : Any
        The data to validate.
    param_name : str
        The name of the parameter being validated.
    function_name : str
        The name of the calling function for error messages.

    Raises
    ------
    TypeError
        If adata is not an AnnData object.
    """
    # Import here to avoid circular dependency
    try:
        from anndata import AnnData
    except ImportError:
        raise ImportError(
            f"[{function_name}] AnnData is required but not installed. "
            "Install with: pip install anndata"
        )

    if not isinstance(adata, AnnData):
        raise TypeError(
            f"[{function_name}] Parameter '{param_name}' must be an AnnData object, "
            f"got {type(adata).__name__}"
        )


# ============================================================================
# Column Validation
# ============================================================================


def validate_column_exists(
    data: pd.DataFrame, column: str, param_name: str, function_name: str
) -> None:
    """
    Validate that a column exists in a DataFrame.

    Parameters
    ----------
    data : pd.DataFrame
        The DataFrame to check.
    column : str
        The column name to validate.
    param_name : str
        The name of the parameter being validated.
    function_name : str
        The name of the calling function for error messages.

    Raises
    ------
    ValueError
        If the column does not exist in the DataFrame.
    """
    if column not in data.columns:
        available = list(data.columns[:10])  # Show first 10 columns
        if len(data.columns) > 10:
            available_str = f"{available}... ({len(data.columns)} total columns)"
        else:
            available_str = str(available)

        raise ValueError(
            f"[{function_name}] Column '{column}' (specified in parameter '{param_name}') "
            f"not found in data. Available columns: {available_str}"
        )


def validate_columns_exist(
    data: pd.DataFrame, columns: List[str], function_name: str
) -> None:
    """
    Validate that multiple columns exist in a DataFrame.

    Parameters
    ----------
    data : pd.DataFrame
        The DataFrame to check.
    columns : List[str]
        The column names to validate.
    function_name : str
        The name of the calling function for error messages.

    Raises
    ------
    ValueError
        If any column does not exist in the DataFrame.
    """
    missing = [col for col in columns if col not in data.columns]
    if missing:
        available = list(data.columns[:10])
        if len(data.columns) > 10:
            available_str = f"{available}... ({len(data.columns)} total columns)"
        else:
            available_str = str(available)

        raise ValueError(
            f"[{function_name}] Column(s) {missing} not found in data. "
            f"Available columns: {available_str}"
        )


# ============================================================================
# AnnData-Specific Validation
# ============================================================================


def validate_adata_layer(adata, layer: str, function_name: str) -> None:
    """
    Validate that a layer exists in an AnnData object.

    Parameters
    ----------
    adata : AnnData
        The AnnData object to check.
    layer : str
        The layer name to validate.
    function_name : str
        The name of the calling function for error messages.

    Raises
    ------
    ValueError
        If the layer does not exist in adata.layers.
    """
    if layer not in adata.layers:
        available = list(adata.layers.keys())
        raise ValueError(
            f"[{function_name}] Layer '{layer}' not found in adata.layers. "
            f"Available layers: {available}"
        )


def validate_adata_obs_columns(
    adata, columns: Union[str, List[str]], function_name: str
) -> None:
    """
    Validate that columns exist in adata.obs.

    Parameters
    ----------
    adata : AnnData
        The AnnData object to check.
    columns : Union[str, List[str]]
        The column name(s) to validate.
    function_name : str
        The name of the calling function for error messages.

    Raises
    ------
    ValueError
        If any column does not exist in adata.obs.
    """
    if isinstance(columns, str):
        columns = [columns]

    missing = [col for col in columns if col not in adata.obs.columns]
    if missing:
        available = list(adata.obs.columns[:10])
        if len(adata.obs.columns) > 10:
            available_str = f"{available}... ({len(adata.obs.columns)} total columns)"
        else:
            available_str = str(available)

        raise ValueError(
            f"[{function_name}] Column(s) {missing} not found in adata.obs. "
            f"Available columns: {available_str}"
        )


def validate_adata_var_columns(
    adata, columns: Union[str, List[str]], function_name: str
) -> None:
    """
    Validate that columns exist in adata.var.

    Parameters
    ----------
    adata : AnnData
        The AnnData object to check.
    columns : Union[str, List[str]]
        The column name(s) to validate.
    function_name : str
        The name of the calling function for error messages.

    Raises
    ------
    ValueError
        If any column does not exist in adata.var.
    """
    if isinstance(columns, str):
        columns = [columns]

    missing = [col for col in columns if col not in adata.var.columns]
    if missing:
        available = list(adata.var.columns[:10])
        if len(adata.var.columns) > 10:
            available_str = f"{available}... ({len(adata.var.columns)} total columns)"
        else:
            available_str = str(available)

        raise ValueError(
            f"[{function_name}] Column(s) {missing} not found in adata.var. "
            f"Available columns: {available_str}"
        )


# ============================================================================
# Data Quality Validation
# ============================================================================


def validate_dataframe_not_empty(data: pd.DataFrame, function_name: str) -> None:
    """
    Validate that a DataFrame is not empty.

    Parameters
    ----------
    data : pd.DataFrame
        The DataFrame to check.
    function_name : str
        The name of the calling function for error messages.

    Raises
    ------
    ValueError
        If the DataFrame is empty.
    """
    if data.empty:
        raise ValueError(
            f"[{function_name}] Data is empty. Cannot create plot with no data."
        )


def validate_no_nulls(
    data: pd.DataFrame,
    columns: Union[str, List[str]],
    function_name: str,
    allow_partial: bool = False,
) -> None:
    """
    Validate that columns have no null values.

    Parameters
    ----------
    data : pd.DataFrame
        The DataFrame to check.
    columns : Union[str, List[str]]
        The column name(s) to validate.
    function_name : str
        The name of the calling function for error messages.
    allow_partial : bool, default False
        If True, only warn if ALL values are null. If False, raise error if ANY null exists.

    Raises
    ------
    ValueError
        If columns contain null values (subject to allow_partial setting).
    """
    if isinstance(columns, str):
        columns = [columns]

    null_counts = {}
    for col in columns:
        null_count = data[col].isnull().sum()
        if null_count > 0:
            null_counts[col] = null_count

    if null_counts:
        if not allow_partial:
            details = ", ".join(
                [f"'{col}': {count}" for col, count in null_counts.items()]
            )
            raise ValueError(
                f"[{function_name}] Null values found in column(s) ({details}). "
                "Remove null values before plotting."
            )
        else:
            # Check if ALL values are null
            all_null_cols = [col for col in columns if data[col].isnull().all()]
            if all_null_cols:
                raise ValueError(
                    f"[{function_name}] Column(s) {all_null_cols} contain only null values. "
                    "At least some valid data is required."
                )


def validate_column_type(
    data: pd.DataFrame, column: str, expected_types: List[str], function_name: str
) -> None:
    """
    Validate that a column has the expected data type.

    Parameters
    ----------
    data : pd.DataFrame
        The DataFrame to check.
    column : str
        The column name to validate.
    expected_types : List[str]
        List of acceptable type names (e.g., ['int', 'float', 'numeric']).
    function_name : str
        The name of the calling function for error messages.

    Raises
    ------
    ValueError
        If the column type is not in expected_types.
    """
    dtype = data[column].dtype

    # Map expected types to pandas dtypes
    numeric_types = ["int", "float", "number", "numeric"]

    if any(t in numeric_types for t in expected_types):
        if not pd.api.types.is_numeric_dtype(dtype):
            raise ValueError(
                f"[{function_name}] Column '{column}' must be numeric, got {dtype}"
            )
    elif (
        "object" in expected_types
        or "string" in expected_types
        or "categorical" in expected_types
    ):
        if not pd.api.types.is_object_dtype(
            dtype
        ) and not pd.api.types.is_categorical_dtype(dtype):
            raise ValueError(
                f"[{function_name}] Column '{column}' must be categorical/string, "
                f"got {dtype}"
            )


def validate_sufficient_data(
    data: pd.DataFrame, column: str, min_count: int, function_name: str
) -> None:
    """
    Validate that a column has sufficient non-null values.

    Parameters
    ----------
    data : pd.DataFrame
        The DataFrame to check.
    column : str
        The column name to validate.
    min_count : int
        Minimum number of non-null values required.
    function_name : str
        The name of the calling function for error messages.

    Raises
    ------
    ValueError
        If the column has fewer than min_count non-null values.
    """
    valid_count = data[column].notna().sum()
    if valid_count < min_count:
        raise ValueError(
            f"[{function_name}] Column '{column}' has only {valid_count} non-null values, "
            f"but at least {min_count} are required."
        )


def validate_binary_column(data: pd.DataFrame, column: str, function_name: str) -> None:
    """
    Validate that a column contains only two unique values.

    Parameters
    ----------
    data : pd.DataFrame
        The DataFrame to check.
    column : str
        The column name to validate.
    function_name : str
        The name of the calling function for error messages.

    Raises
    ------
    ValueError
        If the column does not have exactly 2 unique values.
    """
    unique_values = data[column].dropna().unique()
    if len(unique_values) != 2:
        raise ValueError(
            f"[{function_name}] Column '{column}' must have exactly 2 unique values for binary operations, "
            f"found {len(unique_values)}: {list(unique_values)}"
        )


def validate_categorical_has_levels(
    data: pd.DataFrame,
    column: str,
    min_levels: Optional[int] = None,
    max_levels: Optional[int] = None,
    function_name: str = "",
) -> None:
    """
    Validate that a categorical column has an acceptable number of levels.

    Parameters
    ----------
    data : pd.DataFrame
        The DataFrame to check.
    column : str
        The column name to validate.
    min_levels : Optional[int]
        Minimum number of unique values required.
    max_levels : Optional[int]
        Maximum number of unique values allowed.
    function_name : str
        The name of the calling function for error messages.

    Raises
    ------
    ValueError
        If the number of levels is outside the acceptable range.
    """
    n_levels = data[column].nunique()

    if min_levels is not None and n_levels < min_levels:
        raise ValueError(
            f"[{function_name}] Column '{column}' has {n_levels} unique values, "
            f"but at least {min_levels} are required."
        )

    if max_levels is not None and n_levels > max_levels:
        raise ValueError(
            f"[{function_name}] Column '{column}' has {n_levels} unique values, "
            f"but at most {max_levels} are allowed."
        )


def validate_numeric_range(
    data: pd.DataFrame,
    column: str,
    min_val: Optional[float] = None,
    max_val: Optional[float] = None,
    function_name: str = "",
) -> None:
    """
    Validate that numeric column values are within an acceptable range.

    Parameters
    ----------
    data : pd.DataFrame
        The DataFrame to check.
    column : str
        The column name to validate.
    min_val : Optional[float]
        Minimum acceptable value.
    max_val : Optional[float]
        Maximum acceptable value.
    function_name : str
        The name of the calling function for error messages.

    Raises
    ------
    ValueError
        If any values are outside the acceptable range.
    """
    if min_val is not None:
        if (data[column] < min_val).any():
            raise ValueError(
                f"[{function_name}] Column '{column}' contains values less than {min_val}. "
                "All values must be within the acceptable range."
            )

    if max_val is not None:
        if (data[column] > max_val).any():
            raise ValueError(
                f"[{function_name}] Column '{column}' contains values greater than {max_val}. "
                "All values must be within the acceptable range."
            )


# ============================================================================
# Structure Validation
# ============================================================================


def validate_pairs_format(
    pairs: Any, data: pd.DataFrame, column: str, function_name: str
) -> None:
    """
    Validate that pairs parameter has correct format and values exist in column.

    Parameters
    ----------
    pairs : Any
        The pairs structure to validate (should be list of tuples or "all").
    data : pd.DataFrame
        The DataFrame containing the data.
    column : str
        The column name that pair values should reference.
    function_name : str
        The name of the calling function for error messages.

    Raises
    ------
    ValueError
        If pairs format is invalid or references non-existent values.
    """
    # Allow special value "all"
    if isinstance(pairs, str) and pairs == "all":
        return

    if not isinstance(pairs, list):
        raise ValueError(
            f"[{function_name}] Parameter 'pairs' must be a list of tuples or 'all', "
            f"got {type(pairs).__name__}"
        )

    for i, pair in enumerate(pairs):
        if not isinstance(pair, (tuple, list)) or len(pair) != 2:
            raise ValueError(
                f"[{function_name}] Each pair must be a tuple/list of 2 elements, "
                f"but pair {i} is: {pair}"
            )

        # Check that values exist in the column
        available_values = data[column].unique()
        for val in pair:
            if val not in available_values:
                raise ValueError(
                    f"[{function_name}] Pair value '{val}' not found in column '{column}'. "
                    f"Available values: {list(available_values)}"
                )


def validate_length_match(
    list1: Any, list2: Any, name1: str, name2: str, function_name: str
) -> None:
    """
    Validate that two lists have matching lengths.

    Parameters
    ----------
    list1 : Any
        First list/array.
    list2 : Any
        Second list/array.
    name1 : str
        Name of first parameter.
    name2 : str
        Name of second parameter.
    function_name : str
        The name of the calling function for error messages.

    Raises
    ------
    ValueError
        If the lengths don't match.
    """
    len1 = len(list1) if hasattr(list1, "__len__") else 0
    len2 = len(list2) if hasattr(list2, "__len__") else 0

    if len1 != len2:
        raise ValueError(
            f"[{function_name}] Parameters '{name1}' and '{name2}' must have matching lengths, "
            f"got {len1} and {len2}"
        )


# ============================================================================
# Safe Operations
# ============================================================================


def safe_column_access(
    data: pd.DataFrame, column: str, function_name: str, default: Any = None
) -> Any:
    """
    Safely access a column with a default fallback.

    Parameters
    ----------
    data : pd.DataFrame
        The DataFrame to access.
    column : str
        The column name to access.
    function_name : str
        The name of the calling function for error messages.
    default : Any, default None
        Value to return if column doesn't exist.

    Returns
    -------
    Any
        The column data or default value.

    Raises
    ------
    ValueError
        If column doesn't exist and default is None.
    """
    if column not in data.columns:
        if default is None:
            raise ValueError(
                f"[{function_name}] Column '{column}' not found in data. "
                f"Available columns: {list(data.columns)}"
            )
        return default
    return data[column]


def safe_numeric_conversion(
    value: Any, function_name: str, context: str = ""
) -> Union[int, float]:
    """
    Safely convert a value to numeric, with helpful error on failure.

    Parameters
    ----------
    value : Any
        The value to convert.
    function_name : str
        The name of the calling function for error messages.
    context : str, default ""
        Additional context for the error message.

    Returns
    -------
    Union[int, float]
        The converted numeric value.

    Raises
    ------
    ValueError
        If conversion fails or value is NaN.
    """
    if pd.isna(value):
        context_str = f" ({context})" if context else ""
        raise ValueError(
            f"[{function_name}] Cannot convert NaN to numeric{context_str}. "
            "Ensure all required values are non-null."
        )

    try:
        return int(value) if isinstance(value, (int, np.integer)) else float(value)
    except (ValueError, TypeError) as e:
        context_str = f" ({context})" if context else ""
        raise ValueError(
            f"[{function_name}] Cannot convert value to numeric{context_str}: {value}"
        ) from e


def safe_division(
    numerator: float, denominator: float, default: float = np.nan
) -> float:
    """
    Safely perform division, returning default on division by zero.

    Parameters
    ----------
    numerator : float
        The numerator.
    denominator : float
        The denominator.
    default : float, default np.nan
        Value to return if denominator is zero.

    Returns
    -------
    float
        The result of division or default value.
    """
    if denominator == 0:
        return default
    return numerator / denominator



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from matplotlib.axes import Axes

import sys
from pathlib import Path

import matplotlib as mpl
from matplotlib import font_manager as fm

import cnsplots as cns

_HELVETICA_BOLD_REGISTERED = False


def _ensure_helvetica_bold() -> None:
    """
    Ensure matplotlib can resolve a true Helvetica bold face on macOS.
    """
    global _HELVETICA_BOLD_REGISTERED
    if _HELVETICA_BOLD_REGISTERED:
        return

    for font in fm.fontManager.ttflist:
        if font.name == "Helvetica" and int(font.weight) >= 700:
            _HELVETICA_BOLD_REGISTERED = True
            return

    if sys.platform != "darwin":
        return

    ttc_path = Path("/System/Library/Fonts/Helvetica.ttc")
    if not ttc_path.exists():
        return

    try:
        from fontTools.ttLib import TTCollection
    except Exception:
        return

    bold_ttf_path = Path.home() / ".cache" / "cnsplots" / "fonts" / "Helvetica-Bold.ttf"
    try:
        bold_ttf_path.parent.mkdir(parents=True, exist_ok=True)
        if not bold_ttf_path.exists():
            ttc = TTCollection(str(ttc_path))
            bold_face = None
            for face in ttc.fonts:
                family = (face["name"].getDebugName(1) or "").strip()
                subfamily = (face["name"].getDebugName(2) or "").strip().lower()
                if family == "Helvetica" and "bold" in subfamily:
                    bold_face = face
                    break
            if bold_face is None:
                return
            bold_face.save(str(bold_ttf_path))

        fm.fontManager.addfont(str(bold_ttf_path))
    except Exception:
        return

    for font in fm.fontManager.ttflist:
        if font.name == "Helvetica" and int(font.weight) >= 700:
            _HELVETICA_BOLD_REGISTERED = True
            return


def setup_matplotlib(
    color_cycle: str | None = None,
    color_map: str | None = None,
    fontsize_title: int | float | None = None,
    fontsize_legend: int | float | None = None,
    linewidth_axes: float | None = None,
) -> None:
    """
    Configure matplotlib with publication-quality default styling.

    This function sets up matplotlib with custom styling optimized for
    scientific publications, including font settings, color palettes,
    axis styling, and export parameters.

    Parameters
    ----------
    color_cycle : str, default: None
        Name of the qualitative color palette for the default color cycle.
        If None, uses cns.settings.palette_qual.
        Options include: 'Set1', 'Set2', 'Dark2', 'Ecotyper1'-'Ecotyper6',
        'Tableau', 'Bold', 'ECharts', etc.
    color_map : str, default: None
        Name of the sequential colormap for continuous data.
        If None, uses cns.settings.palette_seq.
        Options include: 'gnuplot', 'parula', 'bwr', 'hot', 'BuRd_custom',
        'WhYlOrRd_custom', etc.
    fontsize_title : int, default: None
        Font size for titles, axis labels, and legend titles.
        If None, uses cns.settings.fontsize_title.
    fontsize_legend : int, default: None
        Font size for tick labels and legend text.
        If None, uses cns.settings.fontsize_legend.
    linewidth_axes : float, default: None
        Line width for axis spines.
        If None, uses cns.settings.linewidth_axes.

    Returns
    -------
    None
        This function modifies matplotlib's global rcParams.

    See Also
    --------
    figure : Initialize a new figure with custom size.
    setup_scanpy : Configure styling for scanpy plots.
    setup_ax : Apply styling to a specific axes.

    Notes
    -----
    The function configures:

    **Fonts:**
    - Family: Sans-serif (Helvetica)
    - Math text: Custom fontset
    - Sizes: title=8pt, legend=7pt (by default)

    **Axes:**
    - Spines: Bottom and left only (top and right hidden)
    - Line width: 0.5pt
    - Colors: Black
    - Margins: 5% on x and y axes
    - Grid: Disabled

    **Ticks:**
    - Size: 2pt length, 0.6pt width
    - Padding: 1pt
    - Colors: Black

    **Legend:**
    - Frame: Disabled
    - Marker scale: 0.5
    - Handle dimensions: 0.7 × 0.7

    **Export:**
    - DPI: 288 (72 × 4)
    - Format: Tight bounding box with 0.01" padding
    - Transparency: Enabled
    - SVG fonts: Embedded as text (fonttype='none')

    **Colormaps:**
    Registers all custom and predefined colormaps for easy access.

    This configuration follows journal guidelines for Cell, Nature, and Science.

    Examples
    --------
    >>> import cnsplots as cns
    >>> # Use default settings
    >>> cns.setup_matplotlib()

    >>> # Customize colors and sizes
    >>> cns.setup_matplotlib(
    ...     color_cycle="Set2", color_map="parula", fontsize_title=10, fontsize_legend=8
    ... )
    """
    # Use settings values if parameters are not provided
    if color_cycle is None:
        color_cycle = cns.settings.palette_qual
    if color_map is None:
        color_map = cns.settings.palette_seq
    if fontsize_title is None:
        fontsize_title = cns.settings.fontsize_title
    if fontsize_legend is None:
        fontsize_legend = cns.settings.fontsize_legend
    if linewidth_axes is None:
        linewidth_axes = cns.settings.linewidth_axes

    _ensure_helvetica_bold()

    def config() -> dict[str, object]:
        """
        Generate matplotlib rcParams configuration dictionary.

        Returns
        -------
        dict
            Configuration dictionary for matplotlib.rcParams.update().
        """
        return {
            "mathtext.fontset": "custom",
            "font.family": "sans-serif",
            "font.sans-serif": [
                "Helvetica",
                "Helvetica Neue",
                "Arial",
                "DejaVu Sans",
                "Liberation Sans",
            ],
            "font.size": fontsize_title,
            "savefig.bbox": "tight",
            "savefig.pad_inches": 0.01,
            "savefig.dpi": 72 * 4,
            "savefig.transparent": True,
            "svg.fonttype": "none",
            # Embed fonts as TrueType (Type 42) in the intermediate PDF so that
            # mutool can resolve every glyph back to a Unicode codepoint.  The
            # default Type 3 encoding has no Unicode map, causing math/symbol
            # glyphs (e.g. the minus sign in 10⁻⁹) to be converted to paths.
            "pdf.fonttype": 42,
            "axes.titlesize": fontsize_title,
            "axes.titleweight": "bold",
            "axes.titlelocation": "center",
            "axes.labelsize": fontsize_title,
            "axes.grid": False,
            "axes.spines.top": False,
            "axes.spines.right": False,
            "axes.linewidth": linewidth_axes,
            "axes.edgecolor": "black",
            "axes.labelcolor": "black",
            "axes.labelpad": 2,
            "axes.titlepad": 4,
            "axes.xmargin": 0.05,
            "axes.ymargin": 0.05,
            "axes.prop_cycle": mpl.cycler(color=cns.palettes(color_cycle)),
            "image.cmap": color_map,
            "legend.fontsize": fontsize_title,
            "legend.title_fontsize": fontsize_title,
            "legend.frameon": False,
            "legend.markerscale": 0.5,
            "legend.handlelength": 0.7,
            "legend.handleheight": 0.7,
            "legend.handletextpad": 0.3,
            "xtick.labelsize": fontsize_legend,
            "xtick.bottom": True,
            "xtick.color": "black",
            "xtick.major.size": 2,
            "xtick.major.width": 0.6,
            "xtick.major.pad": 1,
            "xtick.alignment": "center",
            "ytick.labelsize": fontsize_legend,
            "ytick.left": True,
            "ytick.color": "black",
            "ytick.major.size": 2,
            "ytick.major.width": 0.6,
            "ytick.major.pad": 1,
            "ytick.alignment": "center_baseline",
        }

    for categorical in [
        "Set1",
        "Set2",
        "Set3",
        "Pastel1",
        "Pastel2",
        "Paired",
        "Dark2",
        "Accent",
        "Tableau",
        "Bold",
        "BlueRed",
        "ECharts",
        "Ecotyper1",
        "Ecotyper2",
        "Ecotyper3",
        "Ecotyper4",
        "Ecotyper5",
        "Ecotyper6",
        "NPG",
        "AAAS",
        "Lancet",
        "NEJM",
        "JAMA",
        "JCO",
        "OkabeIto",
        "TolBright",
        "TolMuted",
    ]:
        if categorical not in mpl.colormaps:
            mpl.colormaps.register(
                mpl.colors.ListedColormap(cns.palettes(categorical)), name=categorical
            )
    for continues in [
        "parula",
        "gnuplot",
        "hot",
        "WhYlOrRd_custom",
        "BuRd_custom",
        "OrBu_custom",
        "YlGnBu_custom",
    ]:
        if continues not in mpl.colormaps:
            mpl.colormaps.register(cns.palettes(continues), name=continues)

    mpl.rcParams.update(config())


def setup_scanpy() -> None:
    """
    Configure matplotlib and scanpy with consistent publication-quality styling.

    This function applies cnsplots styling to matplotlib and configures scanpy
    to use compatible figure parameters for single-cell analysis visualizations.

    Parameters
    ----------
    None

    Returns
    -------
    None
        This function modifies global matplotlib and scanpy settings.

    See Also
    --------
    setup_matplotlib : Configure matplotlib styling.
    figure : Initialize a new figure.

    Notes
    -----
    This function performs two configuration steps:

    1. Calls setup_matplotlib() to apply cnsplots styling
    2. Sets scanpy figure parameters:
       - scanpy=False: Disable scanpy's default styling
       - figsize=(2.5, 2.5): Default figure size in inches
       - facecolor='white': White background for figures

    This ensures consistency between cnsplots and scanpy visualizations in
    single-cell analysis workflows.

    Examples
    --------
    >>> import cnsplots as cns
    >>> import scanpy as sc
    >>> cns.setup_scanpy()
    >>> sc.pl.umap(adata, color="cell_type")
    """
    import scanpy

    setup_matplotlib()
    scanpy.set_figure_params(scanpy=False, figsize=(2.5, 2.5), facecolor="white")


def setup_ax(
    ax: Axes,
    fontsize_title: int | float | None = None,
    fontsize_legend: int | float | None = None,
    linewidth_axes: float | None = None,
    colorbar_label: str = "FDR q-val",
) -> None:
    """
    Apply publication-quality styling to a specific matplotlib axes.

    This function formats an existing axes object with consistent styling
    including fonts, spine visibility, tick parameters, and colorbar formatting.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes object to format.
    fontsize_title : int, default: None
        Font size for axis labels and title.
        If None, uses cns.settings.fontsize_title.
    fontsize_legend : int, default: None
        Font size for tick labels and colorbar labels.
        If None, uses cns.settings.fontsize_legend.
    linewidth_axes : float, default: None
        Line width for axis spines.
        If None, uses cns.settings.linewidth_axes.
    colorbar_label : str, default: 'FDR q-val'
        Label for the colorbar (if present). Set to empty string '' to
        remove colorbar label.

    Returns
    -------
    None
        This function modifies the axes in-place.

    See Also
    --------
    setup_matplotlib : Configure global matplotlib styling.
    figure : Initialize a new figure.

    Notes
    -----
    This function applies the following formatting:

    **Font settings:**
    - Family: Sans-serif (Helvetica)
    - Sizes: title=8pt, tick labels=7pt (by default)
    - Color: Black

    **Axis spines:**
    - Visible: Bottom and left only
    - Hidden: Top and right
    - Line width: 0.5pt (by default)
    - Color: Black

    **Ticks:**
    - Size: 2pt length, 0.6pt width
    - Padding: 1pt
    - Color: Black
    - Rotation: 0 degrees

    **Other:**
    - Grid: Disabled
    - Margins: 5% on x and y axes
    - Colorbar: Formatted with matching font settings

    Useful for formatting axes created by external libraries (e.g., seaborn,
    scanpy, gseapy) to match cnsplots styling.

    Examples
    --------
    >>> import cnsplots as cns
    >>> import matplotlib.pyplot as plt
    >>> fig, ax = plt.subplots()
    >>> ax.scatter(x, y)
    >>> cns.setup_ax(ax)

    >>> # Format with custom sizes
    >>> cns.setup_ax(ax, fontsize_title=10, fontsize_legend=9)

    >>> # Format axes from external library
    >>> import seaborn as sns
    >>> ax = sns.heatmap(data)
    >>> cns.setup_ax(ax, colorbar_label="Expression")
    """
    # Use settings values if parameters are not provided
    if fontsize_title is None:
        fontsize_title = cns.settings.fontsize_title
    if fontsize_legend is None:
        fontsize_legend = cns.settings.fontsize_legend
    if linewidth_axes is None:
        linewidth_axes = cns.settings.linewidth_axes

    _ensure_helvetica_bold()

    mpl.rcParams.update(
        {
            "mathtext.fontset": "custom",
            "font.family": "sans-serif",
            "font.sans-serif": [
                "Helvetica",
                "Helvetica Neue",
                "Arial",
                "DejaVu Sans",
                "Liberation Sans",
            ],
        }
    )
    title_props = ax.title.get_fontproperties().copy()
    title_props.set_size(fontsize_title)
    title_props.set_weight("bold")
    ax.set_title(ax.get_title(), fontproperties=title_props, loc="center")
    ax.set_xlabel(ax.get_xlabel(), fontsize=fontsize_title, color="black")
    ax.set_ylabel(ax.get_ylabel(), fontsize=fontsize_title, color="black")
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["bottom"].set_linewidth(linewidth_axes)
    ax.spines["left"].set_linewidth(linewidth_axes)
    ax.spines["bottom"].set_color("black")
    ax.spines["left"].set_color("black")
    ax.tick_params(
        axis="x",
        labelsize=fontsize_legend,
        colors="black",
        length=2,
        width=0.6,
        pad=1,
        labelrotation=0,
    )
    ax.tick_params(
        axis="y",
        labelsize=fontsize_legend,
        colors="black",
        length=2,
        width=0.6,
        pad=1,
        labelrotation=0,
    )
    ax.grid(False)
    ax.margins(x=0.05, y=0.05)
    cbar = ax.collections[0].colorbar if ax.collections else None
    if cbar is not None:
        cbar.ax.tick_params(labelsize=fontsize_legend)
        if colorbar_label:
            cbar.set_label(colorbar_label, fontsize=fontsize_title, color="black")


def setup_ggplot() -> str:
    """
    Generate R ggplot2 theme code matching cnsplots styling.

    This function returns an R code string defining a custom ggplot2 theme
    that matches the cnsplots matplotlib styling, enabling consistent
    visualization across Python and R workflows.

    Parameters
    ----------
    None

    Returns
    -------
    str
        R code string defining a custom ggplot2 theme (theme_custom).

    See Also
    --------
    setup_matplotlib : Configure matplotlib styling.
    setup_scanpy : Configure scanpy styling.

    Notes
    -----
    The returned R code defines a ggplot2 theme with:

    **Fonts:**
    - Size: 10pt for all text elements
    - Family: Sans-serif
    - Face: Plain (no bold or italic)
    - Color: Black

    **Grid:**
    - Major and minor grid lines: Disabled

    **Axes:**
    - Text size: 10pt
    - Title size: 10pt
    - Color: Black

    **Legend:**
    - Text size: 10pt
    - Title size: 10pt

    To use this theme in R:
    1. Copy the returned string to your R script
    2. Apply with: `plot + theme_custom`

    The theme follows the same minimalist design philosophy as cnsplots,
    with clean axes, no grid, and consistent typography.

    Examples
    --------
    >>> import cnsplots as cns
    >>> theme_code = cns.setup_ggplot()
    >>> print(theme_code)

    In R:
    >>> # Copy the theme code, then use:
    >>> p < -ggplot(data, aes(x=x, y=y)) + geom_point()
    >>> p + theme_custom
    """
    return """
    fontsize <- 10
    theme_custom <- theme(
      text = element_text(
        size = fontsize,
        color = 'black',
        family = "sans",
        face = "plain"
      ),
      panel.grid.major = element_blank(),
      panel.grid.minor = element_blank(),

      axis.text.x = element_text(
        size = fontsize,
        color = 'black'
      ),
      axis.text.y = element_text(
        size = fontsize,
        color = 'black'
      ),

      axis.title.x = element_text(
        size = fontsize,
        color = 'black'
      ),
      axis.title.y = element_text(
        size = fontsize,
        color = 'black'
      ),

      axis.title = element_text(
        size = fontsize,
        color = 'black',
        face = "plain"
      ),

      plot.title = element_text(
        size = fontsize,
        color = 'black',
        face = "plain"
      ),

      legend.text = element_text(size = fontsize),
      legend.title = element_text(size = fontsize)
    )
    """



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Union

if TYPE_CHECKING:
    import pandas as pd
    from matplotlib.axes import Axes

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.patches import Patch

import cnsplots as cns
from cnsplots._validation import (
    validate_column_exists,
    validate_columns_exist,
    validate_dataframe,
    validate_dataframe_not_empty,
)

LollipopPair = Union[tuple[str, str], tuple[tuple[str, str], tuple[str, str]]]


def barplot(
    data: pd.DataFrame,
    x: str,
    y: str,
    pairs: list[tuple[str, str]] | None = None,
    addtip: bool = False,
    **kwargs: Any,
) -> Axes:
    """
    Create a bar plot showing mean values across categories with optional statistics.

    This function creates a bar plot displaying the mean of a continuous variable
    grouped by a categorical variable. Error bars are not displayed by default.
    Statistical comparisons and value labels can be added.

    Parameters
    ----------
    data : pd.DataFrame
        The input DataFrame containing the data to be plotted.
    x : str
        Column name for the categorical variable on the x-axis.
    y : str
        Column name for the continuous variable whose means are plotted on the y-axis.
    pairs : list of tuple of str, optional
        List of pairs of category names from x for pairwise statistical comparisons
        using Welch's t-test.
    addtip : bool, default: False
        Whether to add text labels showing the mean value above each bar.
    **kwargs
        Additional keyword arguments passed to `seaborn.barplot`.

    Returns
    -------
    matplotlib.axes.Axes
        The matplotlib Axes object containing the plot.

    See Also
    --------
    boxplot : Create a box plot showing full distribution.
    violinplot : Create a violin plot with distribution shape.
    stackplot : Create a stacked bar plot for categorical data.

    Examples
    --------
    >>> import cnsplots as cns
    >>> ax = cns.barplot(
    ...     data=df,
    ...     x="treatment",
    ...     y="response",
    ...     pairs=[("control", "drug_a"), ("control", "drug_b")],
    ...     addtip=True,
    ... )
    >>> ax.set_title("Treatment Response")

    >>> # Color by group with automatic legend
    >>> ax = cns.barplot(data=df, x="treatment", y="response", palette="cell_type")
    >>> ax.set_ylabel("Mean Response")
    """
    # Validate inputs
    validate_dataframe(data, "data", "barplot")
    validate_columns_exist(data, [x, y], "barplot")
    validate_dataframe_not_empty(data, "barplot")

    args = {
        "edgecolor": None,
        "errorbar": None,
        "err_kws": {"color": "black", "linewidth": 0.7},
        "capsize": 0.2,
    }
    palette = kwargs.pop("palette", None)
    show_legend = False
    legend_handles = []
    if isinstance(palette, str) and palette in data.columns:
        group_col = palette
        unique_labels = data[palette].unique()
        palette_colors = sns.color_palette(n_colors=len(unique_labels))
        label_to_color = dict(zip(unique_labels, palette_colors))
        target_column = palette
        which_numeric = x if not pd.api.types.is_numeric_dtype(data[x]) else y
        mapping_index = (
            data[[which_numeric, target_column]]
            .drop_duplicates()
            .set_index(which_numeric)[target_column]
            .to_dict()
        )
        palette = {k: label_to_color[v] for k, v in mapping_index.items()}
        show_legend = True
        unique_groups = data[group_col].unique()
        color_list = sns.color_palette(n_colors=len(unique_groups))
        group_to_color = dict(zip(unique_groups, color_list))
        legend_handles = [
            Patch(facecolor=color, label=label)
            for label, color in group_to_color.items()
        ]
    plotting = {"data": data, "x": x, "y": y, "palette": palette}
    plotting.update(args)
    plotting.update(kwargs)
    ax = sns.barplot(**plotting)
    if addtip:
        groupedvalues = data.groupby(x)[[y]].mean().reset_index()
        for _, row in groupedvalues.iterrows():
            ax.text(
                row.name,
                row[y] + 0.05,
                round(row[y], 2),
                color="black",
                ha="center",
                rotation=0,
                size=6,
            )
    if pairs is not None:
        cns.utils._p_value_helper("t-test_welch", data, ax, plotting, pairs)
    if show_legend:
        ax.legend(
            handles=legend_handles,
            title=group_col,
            bbox_to_anchor=(1.05, 1),
            loc="upper left",
        )

    return ax


def lollipopplot(
    data: pd.DataFrame,
    x: str,
    y: str,
    hue: str | None = None,
    order: list[str] | None = None,
    hue_order: list[str] | None = None,
    pairs: list[LollipopPair] | None = None,
    addtip: bool = False,
    estimator: str = "mean",
    errorbar: str | None = None,
    markersize: float = 20,
    linewidth: float = 1.5,
    marker: str = "o",
    dodge: float = 0.8,
    color: str | None = None,
    palette: Any = None,
    baseline: float = 0,
) -> Axes:
    """
    Create a lollipop plot showing values across categories.

    A lollipop plot is a cleaner alternative to bar plots, using a stem (line)
    topped with a dot to represent aggregated values for each category.
    Orientation is automatically detected based on which axis is categorical.

    Parameters
    ----------
    data : pd.DataFrame
        The input DataFrame containing the data to be plotted.
    x : str
        Column name for the x-axis variable.
    y : str
        Column name for the y-axis variable.
    hue : str, optional
        Column name for grouping data into separate lollipop sets.
    order : list of str, optional
        Order of categories along the categorical axis.
    hue_order : list of str, optional
        Order of hue levels.
    pairs : list of tuple of str, optional
        List of pairs of category names for pairwise statistical comparisons
        using Welch's t-test.
    addtip : bool, default: False
        Whether to add text labels showing the aggregated value at each dot.
    estimator : {'mean', 'median'}, default: 'mean'
        The aggregation function to compute for each category.
    errorbar : {'se', 'sd', 'ci'}, optional
        Type of error bar to draw. ``'se'`` for standard error, ``'sd'`` for
        standard deviation, ``'ci'`` for 95% confidence interval.
    markersize : float, default: 20
        Size of the dot markers (in points squared, passed to scatter ``s``).
    linewidth : float, default: 1.5
        Width of the stem lines.
    marker : str, default: 'o'
        Marker style for the dots.
    dodge : float, default: 0.8
        Total width for grouped lollipops within each category position.
    color : str, optional
        A single color for all stems and dots (e.g. ``"black"``).
        Overrides ``palette``.
    palette : str, list, or dict, optional
        Colors to use. Can be a seaborn palette name, a list of colors,
        or a column name in data for palette-as-column mapping.
    baseline : float, default: 0
        Value at which the stems originate.

    Returns
    -------
    matplotlib.axes.Axes
        The matplotlib Axes object containing the plot.

    See Also
    --------
    barplot : Create a bar plot showing mean values.
    stripplot : Create a strip plot showing individual data points.
    boxplot : Create a box plot showing full distribution.

    Examples
    --------
    >>> import cnsplots as cns
    >>> ax = cns.lollipopplot(data=df, x="treatment", y="response")
    >>> ax.set_title("Treatment Response")

    >>> # Horizontal lollipop plot
    >>> ax = cns.lollipopplot(data=df, x="response", y="treatment")

    >>> # With hue grouping
    >>> ax = cns.lollipopplot(data=df, x="day", y="total_bill", hue="sex")
    """
    # Validate inputs
    validate_dataframe(data, "data", "lollipopplot")
    validate_columns_exist(data, [x, y], "lollipopplot")
    validate_dataframe_not_empty(data, "lollipopplot")

    # Detect orientation
    x_is_numeric = pd.api.types.is_numeric_dtype(data[x])
    y_is_numeric = pd.api.types.is_numeric_dtype(data[y])
    if x_is_numeric and not y_is_numeric:
        orient = "h"
        cat_col, val_col = y, x
    else:
        orient = "v"
        cat_col, val_col = x, y

    # Category order
    categories = order if order is not None else list(data[cat_col].unique())
    cat_positions = np.arange(len(categories))

    ax = plt.gca()

    # Resolve palette and legend
    show_legend = False
    legend_handles = []
    group_col = None

    if color is not None:
        resolved_colors = [color] * len(categories)
    elif isinstance(palette, str) and palette in data.columns:
        group_col = palette
        unique_labels = data[palette].unique()
        palette_colors = sns.color_palette(n_colors=len(unique_labels))
        label_to_color = dict(zip(unique_labels, palette_colors))
        cat_col_vals = (
            data[[cat_col, palette]]
            .drop_duplicates()
            .set_index(cat_col)[palette]
            .to_dict()
        )
        resolved_colors = [label_to_color[cat_col_vals[c]] for c in categories]
        show_legend = True
        legend_handles = [
            Patch(facecolor=clr, label=label) for label, clr in label_to_color.items()
        ]
    elif palette is not None:
        resolved_colors = sns.color_palette(palette, n_colors=len(categories))
    else:
        resolved_colors = sns.color_palette(n_colors=len(categories))

    agg_func = np.median if estimator == "median" else np.mean

    def _compute_error(group_values, err_type):
        sem = group_values.std() / np.sqrt(len(group_values))
        if err_type == "se":
            return sem
        if err_type == "sd":
            return group_values.std()
        if err_type == "ci":
            return 1.96 * sem
        return 0

    if hue is None:
        # Simple: one stem per category
        grouped = data.groupby(cat_col, sort=False)[val_col]
        means = grouped.agg(agg_func).reindex(categories)

        if orient == "v":
            ax.vlines(
                cat_positions,
                baseline,
                means,
                linewidth=linewidth,
                alpha=0.4,
                colors=resolved_colors,
                zorder=2,
            )
            ax.scatter(
                cat_positions,
                means,
                s=markersize,
                marker=marker,
                c=resolved_colors,
                zorder=3,
            )
            if errorbar is not None:
                for i, cat in enumerate(categories):
                    vals = data.loc[data[cat_col] == cat, val_col]
                    err = _compute_error(vals, errorbar)
                    ax.errorbar(
                        cat_positions[i],
                        means.iloc[i],
                        yerr=err,
                        fmt="none",
                        ecolor="black",
                        elinewidth=0.7,
                        capsize=2,
                    )
            ax.set_xticks(cat_positions)
            ax.set_xticklabels(categories)
            if addtip:
                for i, val in enumerate(means):
                    ax.text(
                        float(cat_positions[i]),
                        val + (means.max() - baseline) * 0.02,
                        f"{val:.2f}",
                        color="black",
                        ha="center",
                        va="bottom",
                        fontsize=6,
                    )
        else:
            ax.hlines(
                cat_positions,
                baseline,
                means,
                linewidth=linewidth,
                alpha=0.4,
                colors=resolved_colors,
                zorder=2,
            )
            ax.scatter(
                means,
                cat_positions,
                s=markersize,
                marker=marker,
                c=resolved_colors,
                zorder=3,
            )
            if errorbar is not None:
                for i, cat in enumerate(categories):
                    vals = data.loc[data[cat_col] == cat, val_col]
                    err = _compute_error(vals, errorbar)
                    ax.errorbar(
                        means.iloc[i],
                        cat_positions[i],
                        xerr=err,
                        fmt="none",
                        ecolor="black",
                        elinewidth=0.7,
                        capsize=2,
                    )
            ax.set_yticks(cat_positions)
            ax.set_yticklabels(categories)
            if addtip:
                for i, val in enumerate(means):
                    ax.text(
                        val + (means.max() - baseline) * 0.02,
                        float(cat_positions[i]),
                        f"{val:.2f}",
                        color="black",
                        ha="left",
                        va="center",
                        fontsize=6,
                    )
    else:
        # Grouped: multiple stems per category
        hue_levels = hue_order if hue_order is not None else list(data[hue].unique())
        n_hue = len(hue_levels)
        width = dodge / n_hue
        offsets = np.linspace(-(dodge - width) / 2, (dodge - width) / 2, n_hue)

        if color is not None:
            hue_colors = [color] * n_hue
        elif palette is not None and not (
            isinstance(palette, str) and palette in data.columns
        ):
            hue_colors = sns.color_palette(palette, n_colors=n_hue)
        else:
            hue_colors = sns.color_palette(n_colors=n_hue)

        for i, hue_val in enumerate(hue_levels):
            subset = data[data[hue] == hue_val]
            grouped = subset.groupby(cat_col, sort=False)[val_col]
            means = grouped.agg(agg_func).reindex(categories)
            positions = cat_positions + offsets[i]
            color = hue_colors[i]

            if orient == "v":
                ax.vlines(
                    positions,
                    baseline,
                    means,
                    linewidth=linewidth,
                    alpha=0.4,
                    colors=color,
                    zorder=2,
                )
                ax.scatter(
                    positions,
                    means,
                    s=markersize,
                    marker=marker,
                    color=color,
                    zorder=3,
                    label=hue_val,
                )
                if errorbar is not None:
                    for j, cat in enumerate(categories):
                        vals = subset.loc[subset[cat_col] == cat, val_col]
                        if len(vals) > 1:
                            err = _compute_error(vals, errorbar)
                            ax.errorbar(
                                positions[j],
                                means.iloc[j],
                                yerr=err,
                                fmt="none",
                                ecolor="black",
                                elinewidth=0.7,
                                capsize=2,
                            )
            else:
                ax.hlines(
                    positions,
                    baseline,
                    means,
                    linewidth=linewidth,
                    alpha=0.4,
                    colors=color,
                    zorder=2,
                )
                ax.scatter(
                    means,
                    positions,
                    s=markersize,
                    marker=marker,
                    color=color,
                    zorder=3,
                    label=hue_val,
                )
                if errorbar is not None:
                    for j, cat in enumerate(categories):
                        vals = subset.loc[subset[cat_col] == cat, val_col]
                        if len(vals) > 1:
                            err = _compute_error(vals, errorbar)
                            ax.errorbar(
                                means.iloc[j],
                                positions[j],
                                xerr=err,
                                fmt="none",
                                ecolor="black",
                                elinewidth=0.7,
                                capsize=2,
                            )

        if orient == "v":
            ax.set_xticks(cat_positions)
            ax.set_xticklabels(categories)
            if addtip:
                for i, hue_val in enumerate(hue_levels):
                    subset = data[data[hue] == hue_val]
                    grouped = subset.groupby(cat_col, sort=False)[val_col]
                    means = grouped.agg(agg_func).reindex(categories)
                    positions = cat_positions + offsets[i]
                    for j, val in enumerate(means):
                        if not np.isnan(val):
                            ax.text(
                                positions[j],
                                val + means.max() * 0.02,
                                f"{val:.2f}",
                                color="black",
                                ha="center",
                                va="bottom",
                                fontsize=6,
                            )
        else:
            ax.set_yticks(cat_positions)
            ax.set_yticklabels(categories)
            if addtip:
                for i, hue_val in enumerate(hue_levels):
                    subset = data[data[hue] == hue_val]
                    grouped = subset.groupby(cat_col, sort=False)[val_col]
                    means = grouped.agg(agg_func).reindex(categories)
                    positions = cat_positions + offsets[i]
                    for j, val in enumerate(means):
                        if not np.isnan(val):
                            ax.text(
                                val + means.max() * 0.02,
                                positions[j],
                                f"{val:.2f}",
                                color="black",
                                ha="left",
                                va="center",
                                fontsize=6,
                            )

    # Statistical annotations
    if pairs is not None:
        # Draw invisible barplot so statannotations can find bar heights
        bar_kw: dict[str, Any] = {
            "data": data,
            "x": x,
            "y": y,
            "ax": ax,
            "alpha": 0,
            "edgecolor": "none",
        }
        if order is not None:
            bar_kw["order"] = order
        if hue is not None:
            bar_kw["hue"] = hue
        if hue_order is not None:
            bar_kw["hue_order"] = hue_order
        sns.barplot(**bar_kw)
        plotting: dict[str, Any] = {"data": data, "x": x, "y": y}
        if order is not None:
            plotting["order"] = order
        if hue is not None:
            plotting["hue"] = hue
        if hue_order is not None:
            plotting["hue_order"] = hue_order
        cns.utils._p_value_helper("t-test_welch", data, ax, plotting, pairs)

    # Legend
    if show_legend:
        ax.legend(
            handles=legend_handles,
            title=group_col,
            bbox_to_anchor=(1.05, 1),
            loc="upper left",
        )

    return ax


def stackplot(
    data: pd.DataFrame,
    x: str,
    y: str,
    bar_order: list[str] | None = None,
    stack_order: list[str] | None = None,
    horizontal: bool = False,
    width: float = 0.5,
    normalize: bool = True,
    pairs: list[tuple[str, str]] | None = None,
    addtip: bool = False,
    n_factor: int | float = 1,
) -> Axes:
    """
    Create a stacked bar plot showing categorical distributions.

    This function creates a stacked bar plot (or horizontal stacked bar plot)
    displaying the distribution or count of one categorical variable across
    levels of another categorical variable.

    Parameters
    ----------
    data : pd.DataFrame
        The input DataFrame containing the data to be plotted.
    x : str
        Column name for the categorical variable determining bar positions.
    y : str
        Column name for the categorical variable determining stack segments.
    bar_order : list, optional
        Order of categories from x to display.
    stack_order : list, optional
        Order of categories from y for stacking.
    horizontal : bool, default: False
        Whether to create horizontal bars instead of vertical bars.
    width : float, default: 0.5
        Width of the bars.
    normalize : bool, default: True
        Whether to normalize counts to frequencies (proportions summing to 1).
    pairs : list of tuple of str, optional
        List of pairs of category names from x for pairwise statistical comparisons.
    addtip : bool, default: False
        Whether to add total count labels above/beside each bar when normalize=True.
    n_factor : int or float, default: 1
        Scaling factor to divide all values.

    Returns
    -------
    matplotlib.axes.Axes
        The matplotlib Axes object containing the plot.

    See Also
    --------
    barplot : Create a bar plot of means.
    pieplot : Create a pie chart for categorical proportions.
    donutplot : Create a donut chart for categorical proportions.

    Examples
    --------
    >>> import cnsplots as cns
    >>> ax = cns.stackplot(
    ...     data=df,
    ...     x="tissue",
    ...     y="cell_type",
    ...     normalize=True,
    ...     pairs=[("lung", "liver")],
    ... )
    >>> ax.set_title("Cell Type Distribution by Tissue")

    >>> # Horizontal stacked bar plot
    >>> ax = cns.stackplot(
    ...     data=df, x="patient", y="mutation", horizontal=True, normalize=False
    ... )
    >>> ax.set_xlabel("Mutation Count")
    """
    # Validate inputs
    validate_dataframe(data, "data", "stackplot")
    validate_columns_exist(data, [x, y], "stackplot")
    validate_dataframe_not_empty(data, "stackplot")

    data2 = data.value_counts([x, y]).reset_index()
    if horizontal:
        df = data2.pivot(index=y, columns=x, values="count")
    else:
        df = data2.pivot(index=x, columns=y, values="count")
    contingency = df.copy()
    if stack_order is not None:
        df.columns = pd.CategoricalIndex(
            df.columns.values, ordered=True, categories=stack_order, name=y
        )
        df = df.sort_index(axis=1)
    if normalize:
        df = df.div(df.sum(axis=1), axis=0)
        value_label = "Frequency"
    else:
        value_label = "Count"
    df = df.reindex(index=bar_order)
    df = df / n_factor
    ax = plt.gca()
    if horizontal:
        ax = df.plot.barh(stacked=True, width=width, ax=ax, rot=0)
        ax.set_ylabel("")
        ax.set_xlabel(value_label)
    else:
        ax = df.plot.bar(stacked=True, width=width, ax=ax, rot=0)
        ax.set_ylabel(value_label)
        ax.set_xlabel("")
    cns.take_legend_out()
    if addtip and normalize:
        tips = (
            contingency.sum(axis=1).astype(int).reindex(index=bar_order).reset_index()
        )
        tips = tips.rename(columns={0: "value"})
        for _, row in tips.iterrows():
            if horizontal:
                ax.text(
                    1 + 0.02,
                    row.name,
                    row["value"],
                    color="black",
                    ha="left",
                    va="center",
                    rotation=0,
                    size=6,
                )
            else:
                ax.text(
                    row.name,
                    1 + 0.02,
                    row["value"],
                    color="black",
                    ha="center",
                    rotation=0,
                    size=6,
                )
    if pairs is not None:
        if horizontal:
            plotting = {"data": data2, "x": "count", "y": y, "order": bar_order}
        else:
            plotting = {"data": data2, "x": x, "y": "count", "order": bar_order}
        if contingency.shape[1] == 2:
            cns.utils._p_value_helper(
                "fisher-exact", data2, ax, plotting, pairs, contingency
            )
        else:
            cns.utils._p_value_helper(
                "chi-squared", data2, ax, plotting, pairs, contingency
            )

    return ax


def stripplot(
    data: pd.DataFrame,
    x: str,
    y: str,
    size: float = 2,
    showmedian: bool = True,
    showmeans: bool = False,
    addcount: bool = False,
    **kwargs: Any,
) -> Axes:
    """
    Create a strip plot showing individual data points with optional summary statistics.

    This function creates a categorical scatter plot (strip plot) where all individual
    data points are displayed, optionally overlaid with median or mean markers.

    Parameters
    ----------
    data : pd.DataFrame
        The input DataFrame containing the data to be plotted.
    x : str
        Column name for the categorical variable on the x-axis.
    y : str
        Column name for the continuous variable on the y-axis.
    size : float, default: 2
        Size of individual data point markers.
    showmedian : bool, default: True
        Whether to overlay a horizontal line at the median for each category.
    showmeans : bool, default: False
        Whether to overlay a marker at the mean for each category.
    addcount : bool, default: False
        Whether to add sample size (n) labels above each category.
    **kwargs
        Additional keyword arguments passed to `seaborn.stripplot`.

    Returns
    -------
    matplotlib.axes.Axes
        The matplotlib Axes object containing the plot.

    See Also
    --------
    boxplot : Create a box plot showing quartiles.
    violinplot : Create a violin plot showing distribution shape.
    scatterplot : Create a general scatter plot.

    Examples
    --------
    >>> import cnsplots as cns
    >>> ax = cns.stripplot(
    ...     data=df, x="treatment", y="response", showmedian=True, addcount=True
    ... )
    >>> ax.set_title("Treatment Response")

    >>> # With grouping and means
    >>> ax = cns.stripplot(
    ...     data=df, x="tissue", y="expression", hue="genotype", showmeans=True, size=3
    ... )
    >>> ax.set_ylabel("Gene Expression")
    """
    # Validate inputs
    validate_dataframe(data, "data", "stripplot")
    validate_columns_exist(data, [x, y], "stripplot")
    validate_dataframe_not_empty(data, "stripplot")

    ax = sns.stripplot(data=data, x=x, y=y, size=size, **kwargs)
    sns.boxplot(
        data=data,
        x=x,
        y=y,
        medianprops={"visible": showmedian, "color": "black", "lw": 1},
        meanprops={
            "markerfacecolor": "white",
            "markeredgecolor": "black",
            "marker": "o",
            "markersize": size + 1,
        },
        whiskerprops={"visible": False},
        zorder=10,
        showfliers=False,
        showbox=False,
        showcaps=False,
        width=0.3,
        ax=ax,
        showmeans=showmeans,
    )
    if addcount:
        cns.utils._addcount_helper(data, x, ax)

    if ax.get_legend() is not None:
        for handle in ax.get_legend().legend_handles:
            if hasattr(handle, "set_sizes"):
                handle.set_sizes([size**2])
            elif hasattr(handle, "set_markersize"):
                handle.set_markersize(size * 2)

    return ax


def pieplot(
    data: pd.DataFrame,
    x: str,
    legend: str = "bottom",
    hue_order: list[str] | None = None,
) -> Axes:
    """
    Create a pie chart showing categorical proportions.

    Parameters
    ----------
    data : pd.DataFrame
        The input DataFrame containing the data to be plotted.
    x : str
        Column name for the categorical variable to visualize.
    legend : {'right', 'left', 'top', 'bottom'}, default: 'bottom'
        Position of the legend relative to the pie chart.
    hue_order : list, optional
        Order of categories to display in the pie chart.

    Returns
    -------
    matplotlib.axes.Axes
        The matplotlib Axes object containing the plot.

    See Also
    --------
    donutplot : Create a donut chart (pie chart with center hole).
    stackplot : Create a stacked bar plot for categorical distributions.
    barplot : Create a bar plot showing category frequencies.

    Examples
    --------
    >>> import cnsplots as cns
    >>> ax = cns.pieplot(data=df, x="cell_type")
    >>> ax.set_title("Cell Type Distribution")

    >>> # Specify category order
    >>> ax = cns.pieplot(data=df, x="response", hue_order=["CR", "PR", "SD", "PD"])
    """
    # Validate inputs
    validate_dataframe(data, "data", "pieplot")
    validate_column_exists(data, x, "x", "pieplot")
    validate_dataframe_not_empty(data, "pieplot")

    df = data[x].value_counts()
    if hue_order is None:
        hue_order = df.index
    ax = plt.gca()
    ax = df.reindex(index=hue_order).plot.pie(
        shadow=False,
        autopct="%1.0f%%",
        explode=[0] * df.shape[0],
        textprops={"fontsize": 6, "color": "white"},
        labeldistance=None,
        wedgeprops={"linewidth": 0.3, "edgecolor": "white"},
        ax=ax,
        ylabel="",
        legend=True,
    )
    legend_positions = {
        "right": {"loc": "upper left", "bbox_to_anchor": (1, 1.02)},
        "left": {"loc": "upper right", "bbox_to_anchor": (-0.02, 1.02)},
        "top": {"loc": "lower center", "bbox_to_anchor": (0.5, 1.05)},
        "bottom": {"loc": "upper center", "bbox_to_anchor": (0.5, -0.05)},
    }
    pos = legend_positions.get(legend, legend_positions["right"])
    ax.legend(**pos, title=x)
    return ax


def donutplot(
    data: pd.DataFrame,
    x: str,
    legend: str = "bottom",
    hue_order: list[str] | None = None,
) -> Axes:
    """
    Create a donut chart showing categorical proportions.

    Parameters
    ----------
    data : pd.DataFrame
        The input DataFrame containing the data to be plotted.
    x : str
        Column name for the categorical variable to visualize.
    legend : {'right', 'left', 'top', 'bottom'}, default: 'bottom'
        Position of the legend relative to the pie chart.
    hue_order : list, optional
        Order of categories to display in the donut chart.

    Returns
    -------
    matplotlib.axes.Axes
        The matplotlib Axes object containing the plot.

    See Also
    --------
    pieplot : Create a pie chart without center hole.
    stackplot : Create a stacked bar plot for categorical distributions.
    vennplot : Create a Venn diagram for set overlaps.

    Examples
    --------
    >>> import cnsplots as cns
    >>> ax = cns.donutplot(data=df, x="tissue_type")
    >>> ax.set_title("Tissue Type Distribution")

    >>> # Specify category order
    >>> ax = cns.donutplot(data=df, x="grade", hue_order=["I", "II", "III", "IV"])
    """
    # Validate inputs
    validate_dataframe(data, "data", "donutplot")
    validate_column_exists(data, x, "x", "donutplot")
    validate_dataframe_not_empty(data, "donutplot")

    df = data[x].value_counts()
    if hue_order is None:
        hue_order = df.index
    ax = plt.gca()
    ax = df.reindex(index=hue_order).plot.pie(
        labeldistance=None,
        ax=ax,
        ylabel="",
        legend=True,
        wedgeprops={"edgecolor": "black", "linewidth": 0.3},
    )
    ax.add_patch(
        plt.Circle(
            (0, 0), radius=0.6, facecolor="white", edgecolor="black", linewidth=0.3
        )
    )
    plt.annotate(x, (0, 0), size=7, ha="center", va="center")
    cns.utils._remove_edge_from_legend_items(ax)
    legend_positions = {
        "right": {"loc": "upper left", "bbox_to_anchor": (1, 1.02)},
        "left": {"loc": "upper right", "bbox_to_anchor": (-0.02, 1.02)},
        "top": {"loc": "lower center", "bbox_to_anchor": (0.5, 1.05)},
        "bottom": {"loc": "upper center", "bbox_to_anchor": (0.5, -0.05)},
    }
    pos = legend_positions.get(legend, legend_positions["right"])
    ax.legend(**pos, title=x)
    return ax



from __future__ import annotations

import builtins
from collections.abc import Mapping, Sequence
import sys
import types
from pathlib import Path
from typing import Any

import anndata as ad
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest
from matplotlib.patches import PathPatch

import cnsplots as cns
from cnsplots import _methods, _setup, _svg, _utils, _validation
from cnsplots.helpers import _heatmap as helper_heatmap, _phylo, _sankey
from cnsplots.plots import _distribution as dist_mod
from cnsplots.plots import _genomics as genomics_mod
from cnsplots.plots import _heatmap as heatmap_mod
from cnsplots.plots import _specialized as specialized_mod


def test_methods_gap_coverage(monkeypatch: pytest.MonkeyPatch) -> None:
    class FakeCoxPHFitter:
        def __init__(self) -> None:
            self.summary = pd.DataFrame(
                {
                    "exp(coef)": [1.0],
                    "exp(coef) lower 95%": [0.9],
                    "exp(coef) upper 95%": [1.1],
                    "p": [0.5],
                    "covariate": ["123"],
                },
                index=pd.Index(["123"]),
            )

        def fit(
            self, data: pd.DataFrame, duration_col: str, event_col: str, formula: str
        ) -> None:
            return None

    monkeypatch.setattr(_methods.ll, "CoxPHFitter", FakeCoxPHFitter)
    model = cns.CoxModel(
        pd.DataFrame({"time": [1, 2], "event": [1, 0], "x": [0, 1]}),
        duration="time",
        event="event",
        variates=["123"],
    )
    model.fit()
    assert model.results is not None
    assert model.results["display_label"].isna().all()

    samples = [np.array([0, 0, 0, 0]), np.array([0, 1, 2, 3])]

    def fake_choice(n: int, size: int, replace: bool = True) -> np.ndarray:
        return samples.pop(0)

    monkeypatch.setattr(_methods.np.random, "choice", fake_choice)
    logistic = cns.LogisticModel(
        pd.DataFrame({"event": [0, 0, 1, 1]}), event="event", variates=["1"]
    )
    auc, lower, upper = logistic._compute_auc_ci(
        np.array([0, 0, 1, 1]),
        np.array([0.1, 0.2, 0.8, 0.9]),
        n_bootstrap=2,
    )
    assert lower <= auc <= upper


def test_multipanel_gap_coverage(monkeypatch: pytest.MonkeyPatch) -> None:
    mp = cns.multipanel()
    mp._panels = [{"label": "A", "_below": "missing"}]
    assert mp._get_panel_position(0) == (0, 0)

    mp2 = cns.multipanel()
    mp2._panels = [
        {
            "label": "A",
            "width": 10,
            "height": 10,
            "label_left": 0,
            "label_top": 0,
            "pad_left": 0,
            "pad_top": 0,
            "margin_left": 0,
            "margin_top": 0,
            "margin_right": 0,
            "margin_bottom": 0,
        }
    ]
    assert mp2._get_panel_position(0) == (0, 0)

    mp3 = cns.multipanel(max_width=200)
    mp3.panel(None, width=40, height=20)
    mp3.panel("B", width=40, height=20)
    assert mp3._get_panel_position(1)[0] > 0

    mp4 = cns.multipanel()
    monkeypatch.setattr(mp4, "_create_or_update_figure", lambda: None)
    with pytest.raises(RuntimeError, match="axes was not created"):
        mp4.panel("A", width=10, height=10)


def test_setup_gap_coverage(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    real_import = builtins.__import__
    real_exists = Path.exists

    def fake_exists(self: Path) -> bool:
        if str(self) == "/System/Library/Fonts/Helvetica.ttc":
            return True
        if str(self).endswith("Helvetica-Bold.ttf"):
            return False
        return real_exists(self)

    monkeypatch.setattr(Path, "exists", fake_exists)
    monkeypatch.setattr(_setup, "_HELVETICA_BOLD_REGISTERED", False)
    monkeypatch.setattr(_setup.sys, "platform", "darwin")
    monkeypatch.setattr(_setup.Path, "home", lambda: tmp_path)
    monkeypatch.setattr(
        _setup.fm,
        "fontManager",
        types.SimpleNamespace(ttflist=[], addfont=lambda path: None),
    )

    def missing_fonttools(
        name: str,
        globals: Mapping[str, object] | None = None,
        locals: Mapping[str, object] | None = None,
        fromlist: Sequence[str] = (),
        level: int = 0,
    ) -> object:
        if name == "fontTools.ttLib":
            raise ImportError("missing")
        return real_import(name, globals, locals, fromlist, level)

    monkeypatch.setattr(builtins, "__import__", missing_fonttools)
    _setup._ensure_helvetica_bold()

    class FakeNameTable:
        def __init__(self, family: str, subfamily: str) -> None:
            self.family = family
            self.subfamily = subfamily

        def getDebugName(self, index: int) -> str:
            return self.family if index == 1 else self.subfamily

    class FakeRegularFace(dict):
        def __init__(self) -> None:
            super().__init__(name=FakeNameTable("Helvetica", "Regular"))

        def save(self, path: str) -> None:
            raise AssertionError("regular face should not be saved")

    class NoBoldCollection:
        def __init__(self, path: str) -> None:
            self.fonts = [FakeRegularFace()]

    def import_fonttools_without_bold(
        name: str,
        globals: Mapping[str, object] | None = None,
        locals: Mapping[str, object] | None = None,
        fromlist: Sequence[str] = (),
        level: int = 0,
    ) -> object:
        if name == "fontTools.ttLib":
            return types.SimpleNamespace(TTCollection=NoBoldCollection)
        return real_import(name, globals, locals, fromlist, level)

    monkeypatch.setattr(builtins, "__import__", import_fonttools_without_bold)
    _setup._ensure_helvetica_bold()

    class FakeFace(dict):
        def __init__(self) -> None:
            super().__init__(name=FakeNameTable("Helvetica", "Bold"))
            self.saved = False

        def save(self, path: str) -> None:
            self.saved = True

    face = FakeFace()

    class FakeCollection:
        def __init__(self, path: str) -> None:
            self.fonts = [face]

    font_manager = types.SimpleNamespace(
        ttflist=[],
        addfont=lambda path: font_manager.ttflist.append(
            types.SimpleNamespace(name="Helvetica", weight=700)
        ),
    )
    monkeypatch.setattr(_setup.fm, "fontManager", font_manager)

    def import_fonttools(
        name: str,
        globals: Mapping[str, object] | None = None,
        locals: Mapping[str, object] | None = None,
        fromlist: Sequence[str] = (),
        level: int = 0,
    ) -> object:
        if name == "fontTools.ttLib":
            return types.SimpleNamespace(TTCollection=FakeCollection)
        return real_import(name, globals, locals, fromlist, level)

    monkeypatch.setattr(builtins, "__import__", import_fonttools)
    _setup._ensure_helvetica_bold()
    assert _setup._HELVETICA_BOLD_REGISTERED is True
    assert face.saved is True

    monkeypatch.setattr(_setup, "_HELVETICA_BOLD_REGISTERED", False)
    bad_font_manager = types.SimpleNamespace(
        ttflist=[], addfont=lambda path: (_ for _ in ()).throw(RuntimeError("boom"))
    )
    monkeypatch.setattr(_setup.fm, "fontManager", bad_font_manager)
    _setup._ensure_helvetica_bold()


def test_svg_gap_coverage() -> None:
    class FakeText:
        def getparent(self) -> None:
            return None

        def xpath(self, pattern: str, namespaces: dict[str, str]) -> list[object]:
            return []

    class FakeRoot:
        def xpath(self, pattern: str, namespaces: dict[str, str]) -> list[object]:
            return [FakeText()]

    _svg._process_text_elements_lxml(FakeRoot(), {"svg": "http://www.w3.org/2000/svg"})

    class FakeGroup:
        def getparent(self) -> None:
            return None

    class FakeGroupRoot:
        def xpath(self, pattern: str, namespaces: dict[str, str]) -> list[object]:
            return [FakeGroup()]

    _svg._flatten_groups(FakeGroupRoot(), {"svg": "http://www.w3.org/2000/svg"})
    _svg._restore_bold_fonts(
        types.SimpleNamespace(), {"svg": "http://www.w3.org/2000/svg"}, set()
    )


def test_utils_gap_coverage(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    cns.figure(100, 100)
    plt.plot([0, 1], [0, 1])
    nested_path = tmp_path / "nested" / "plot.png"
    cns.savefig(str(nested_path))
    assert nested_path.exists()

    monkeypatch.chdir(tmp_path)
    cns.figure(100, 100)
    plt.plot([0, 1], [0, 1])
    cns.savefig("plot.png")
    assert Path("plot.png").exists()

    fig, ax = plt.subplots()
    plt.sca(ax)
    _utils.take_legend_out()
    assert ax.get_legend().get_title().get_text() == ""

    assert _utils._is_qualitative_cmap(["a"]) is True
    assert _utils._is_qualitative_cmap({"a": "#111111"}) is True

    plt.figure()
    plt.title("β")
    _utils.apply_unicode_font()

    class DummyResult:
        test_short_name = "T"
        significance_suffix = ""

        def __init__(self, pvalue: float) -> None:
            self.pvalue = pvalue

    class DummyAnnotator:
        def __init__(self, ax: object, pairs: object, **plotting: object) -> None:
            self._pvalue_format = None

        def configure(self, **kwargs: object) -> None:
            return None

        def apply_and_annotate(self) -> None:
            return None

        def set_pvalues(self, pvalues: list[float]) -> None:
            return None

        def annotate(self) -> None:
            return None

    monkeypatch.setattr(_utils, "Annotator", DummyAnnotator)
    formatter_ax = plt.subplots()[1]
    _utils._p_value_helper(
        "Mann-Whitney",
        pd.DataFrame({"g": ["A", "B"], "v": [1, 2]}),
        formatter_ax,
        {"x": "g", "y": "v"},
        "all",
        format="full",
    )
    formatter = DummyAnnotator(formatter_ax, [])._pvalue_format
    formatter = formatter or _utils.Annotator(formatter_ax, [])._pvalue_format

    class CaptureAnnotator(DummyAnnotator):
        last = None

        def __init__(self, ax: object, pairs: object, **plotting: object) -> None:
            super().__init__(ax, pairs, **plotting)
            CaptureAnnotator.last = self

    monkeypatch.setattr(_utils, "Annotator", CaptureAnnotator)
    fig2, ax2 = plt.subplots()
    _utils._p_value_helper(
        "Mann-Whitney",
        pd.DataFrame({"g": ["A", "B"], "v": [1, 2]}),
        ax2,
        {"x": "g", "y": "v"},
        "all",
        format="full",
    )
    formatted = CaptureAnnotator.last._pvalue_format.format_data(DummyResult(0.2))
    assert formatted == "ns"

    with pytest.raises(ValueError, match="requires a hue column"):
        _utils._p_value_helper(
            "Mann-Whitney",
            pd.DataFrame({"group": ["A", "B"], "value": [1, 2]}),
            ax2,
            {"x": "value", "y": "group"},
            "hue",
        )

    for name in ["Set3", "Pastel1", "Pastel2", "Paired", "Dark2", "Accent", "Bold"]:
        assert _utils.palettes(name)


def test_validation_gap_coverage(heatmap_adata: ad.AnnData) -> None:
    wide_df = pd.DataFrame(
        np.zeros((1, 11)), columns=pd.Index([f"c{i}" for i in range(11)])
    )
    with pytest.raises(ValueError, match="11 total columns"):
        _validation.validate_column_exists(wide_df, "missing", "x", "func")
    with pytest.raises(ValueError, match="11 total columns"):
        _validation.validate_columns_exist(wide_df, ["missing"], "func")

    adata = heatmap_adata.copy()
    for i in range(11):
        adata.obs[f"obs{i}"] = i
        adata.var[f"var{i}"] = i
    with pytest.raises(ValueError, match="13 total columns"):
        _validation.validate_adata_obs_columns(adata, "missing", "func")
    with pytest.raises(ValueError, match="13 total columns"):
        _validation.validate_adata_var_columns(adata, "missing", "func")
    with pytest.raises(ValueError, match="Null values found"):
        _validation.validate_no_nulls(pd.DataFrame({"a": [1, None]}), "a", "func")


def test_helper_heatmap_gap_coverage(monkeypatch: pytest.MonkeyPatch) -> None:
    calls: list[object] = []

    def fake_plot(self: object) -> None:
        self.ax = plt.gca()
        self.ax_heatmap = plt.gca()
        self.yticklabels = []

    monkeypatch.setattr(helper_heatmap.ClusterMapPlotterNew, "plot", fake_plot)
    monkeypatch.setattr(
        helper_heatmap.ClusterMapPlotterNew, "post_processing", lambda self: None
    )
    monkeypatch.setattr(
        helper_heatmap.ClusterMapPlotterNew,
        "plot_legends",
        lambda self, ax=None: calls.append(ax),
    )

    annotation = types.SimpleNamespace(
        plot_legend=False,
        legend_list=[],
        label_max_width=0,
        collect_legends=lambda: None,
    )
    cns.figure(100, 100)
    plotter = helper_heatmap.ClusterMapPlotterNew(
        data=pd.DataFrame([[1]]),
        right_annotation=annotation,
        plot=True,
        plot_legend=True,
        legend_anchor="auto",
        verbose=0,
    )
    assert calls[-1] is plotter.ax

    cns.figure(100, 100)
    plotter2 = helper_heatmap.ClusterMapPlotterNew(
        data=pd.DataFrame([[1]]),
        plot=True,
        plot_legend=True,
        legend_anchor="auto",
        verbose=0,
    )
    assert calls[-1] is plotter2.ax_heatmap

    cns.figure(100, 100)
    plotter3 = helper_heatmap.ClusterMapPlotterNew(
        data=pd.DataFrame([[0, 1], [1, 0]]),
        cmap=["red", "blue"],
        plot=False,
        verbose=1,
    )
    plotter3.data = np.array([[0, 1], [1, 0]])
    plotter3.ax = plt.gca()
    plotter3.yticklabels = []
    plotter3.collect_legends()
    assert plotter3.legend_list

    cns.figure(100, 100)
    plotter4 = helper_heatmap.ClusterMapPlotterNew(
        data=pd.DataFrame([[0, 1], [1, 0]]),
        cmap={"0": "red", "1": "blue"},
        plot=False,
        verbose=0,
    )
    plotter4.data = np.array([[0, 1], [1, 0]])
    plotter4.ax = plt.gca()
    plotter4.yticklabels = []
    plotter4.collect_legends()
    assert plotter4.legend_list

    cns.figure(100, 100)
    plotter5 = helper_heatmap.ClusterMapPlotterNew(
        data=pd.DataFrame([[0, 1], [1, 0]]),
        plot=False,
        verbose=0,
    )
    plotter5.ax = plt.gca()
    plotter5.widths = [1, 1, 1]
    plotter5.heights = [1, 1, 1]
    gs = plt.gcf().add_gridspec(1, 1)
    plotter5._define_axes(gs[0])


def test_phylo_and_sankey_gap_coverage() -> None:
    fig, ax = plt.subplots()
    out_ax, cmap = _phylo._heatmap(
        np.array([["A"], ["B"]]),
        cmap={"A": "red"},
        ax=ax,
        leg_pos="right",
    )
    assert out_ax is ax
    assert set(cmap) == {"A", "B"}

    assert len(_phylo._gen_colors(["red", "blue"], 2)) == 2
    assert _phylo._is_categorical(np.array([[1, 2], [3, 4]])) == [False, False]
    assert _phylo._is_categorical("plain text") is False

    def fail_palette(name: str, n: int) -> list[object]:
        raise ValueError("fallback")

    monkeypatch = pytest.MonkeyPatch()
    monkeypatch.setattr(_phylo.sns, "color_palette", fail_palette)
    try:
        assert len(_phylo._gen_colors("viridis", 2)) == 2
    finally:
        monkeypatch.undo()

    _sankey.check_data_matches_labels({"A", "B"}, {"A", "B"}, "left")
    _sankey.check_data_matches_labels(["A", "B"], ["A", "B"], "left")
    data_frame = pd.DataFrame(
        {
            "left": ["A", "B"],
            "right": ["C", "D"],
            "leftWeight": [1, 1],
            "rightWeight": [1, 1],
        }
    )
    _sankey.identify_labels(data_frame, ["A", "B"], ["C", "D"])
    fig2, ax2 = plt.subplots()
    widths_left, _ = _sankey._get_positions_and_total_widths(
        data_frame, ["A", "B"], "left"
    )
    widths_right, _ = _sankey._get_positions_and_total_widths(
        data_frame, ["C", "D"], "right"
    )
    assert set(widths_left) == {"A", "B"}
    assert set(widths_right) == {"C", "D"}
    _sankey.plot_strips(
        ax2,
        {"A": "red", "B": "blue", "C": "green", "D": "black"},
        data_frame,
        ["A", "B"],
        {"A": {"bottom": 0, "left": 1}, "B": {"bottom": 1.04, "left": 1}},
        {"A": {"C": 1, "D": 0}, "B": {"C": 0, "D": 1}},
        {"A": {"C": 1, "D": 0}, "B": {"C": 0, "D": 1}},
        True,
        ["C", "D"],
        {"C": {"bottom": 0, "right": 1}, "D": {"bottom": 1.04, "right": 1}},
        np.float64(1.0),
    )


def test_plot_gap_coverage(
    categorical_df: pd.DataFrame,
    stack_df: pd.DataFrame,
    heatmap_adata: ad.AnnData,
    confusion_df: pd.DataFrame,
    survival_df: pd.DataFrame,
    phylo_adata: ad.AnnData,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cns.figure(120, 120)
    cns.lollipopplot(categorical_df, x="group", y="value", addtip=True, errorbar="bad")

    cns.figure(120, 120)
    cns.lollipopplot(
        categorical_df,
        x="group",
        y="value",
        hue="hue",
        color="black",
        order=["A", "B", "C"],
        hue_order=["H1", "H2"],
        pairs=[(("A", "H1"), ("A", "H2"))],
    )

    cns.figure(120, 120)
    cns.lollipopplot(
        categorical_df.rename(columns={"group": "cat", "value": "num"}),
        x="num",
        y="cat",
        hue="hue",
        addtip=True,
        errorbar="se",
    )

    cns.figure(120, 120)
    cns.stackplot(
        stack_df,
        x="treatment",
        y="response",
        horizontal=True,
        normalize=True,
        addtip=True,
        bar_order=["A", "B", "C"],
    )

    class SizeHandle:
        def __init__(self) -> None:
            self.sizes = None

        def set_sizes(self, sizes: list[float]) -> None:
            self.sizes = sizes

    fake_legend = types.SimpleNamespace(legend_handles=[SizeHandle()])
    monkeypatch.setattr(
        type(plt.gca()),
        "get_legend",
        lambda self: fake_legend,
        raising=False,
    )

    cns.figure(120, 120)
    cns.stripplot(categorical_df, x="group", y="value", hue="hue")
    cns.figure(120, 120)
    cns.scatterplot(
        categorical_df.rename(columns={"value": "x"}).assign(y=lambda df: df["x"] * 2),
        x="x",
        y="y",
        hue="hue",
    )
    cns.figure(120, 120)
    cns.regplot(
        categorical_df.rename(columns={"value": "x"}).assign(
            y=lambda df: df["x"] * 2, color_col=["A"] * 12
        ),
        x="x",
        y="y",
        color="color_col",
    )

    monkeypatch.undo()

    cns.figure(120, 120)
    cns.pieplot(categorical_df, x="group")
    cns.figure(120, 120)
    cns.donutplot(categorical_df, x="group")

    class FakeAxes:
        def __init__(self) -> None:
            self.patches: list[PathPatch] = []
            self.artists = [
                types.SimpleNamespace(
                    get_facecolor=lambda: (1, 0, 0, 1),
                    set_edgecolor=lambda x: None,
                    set_facecolor=lambda x: None,
                )
            ]
            self.lines = [
                types.SimpleNamespace(
                    set_color=lambda x: None,
                    set_mfc=lambda x: None,
                    set_mec=lambda x: None,
                )
            ] * 5

        def get_legend_handles_labels(self) -> tuple[list[object], list[str]]:
            return [], []

        def legend(self, handles: list[object], labels: list[str]) -> None:
            return None

    monkeypatch.setattr(dist_mod.sns, "boxplot", lambda **kwargs: FakeAxes())
    monkeypatch.setattr(cns.utils, "_remove_edge_from_legend_items", lambda ax: None)
    cns.boxplot(categorical_df, x="group", y="value")

    cns.figure(120, 120)
    cns.histplot(
        data=categorical_df.rename(columns={"value": "x", "group": "y"}), y="x"
    )

    adata_nan = heatmap_adata.copy()
    adata_nan.obs["cluster"] = pd.Series(
        ["A", None, "B"], index=adata_nan.obs_names, dtype=object
    )
    cns.figure(120, 120)
    cns.heatmapplot(adata_nan, row_annotation=["cluster"], cmap=None)

    with pytest.raises(ValueError, match="required cell for stats"):
        cns.confusionplot(
            confusion_df,
            x="pred",
            y="truth",
            add_pvalue=True,
            x_order=["neg", "pos"],
            y_order=["neg", "pos"],
            positive_x="missing",
        )

    original_crosstab = heatmap_mod.pd.crosstab
    monkeypatch.setattr(
        heatmap_mod.pd,
        "crosstab",
        lambda *args, **kwargs: pd.DataFrame(
            [[np.nan, 1], [1, 1]],
            index=pd.Index(["neg", "pos"]),
            columns=pd.Index(["neg", "pos"]),
        ),
    )
    with pytest.raises(ValueError, match="TN cell contains NaN"):
        cns.confusionplot(
            confusion_df,
            x="pred",
            y="truth",
            add_pvalue=True,
            annot=False,
            x_order=["neg", "pos"],
            y_order=["neg", "pos"],
        )
    monkeypatch.setattr(heatmap_mod.pd, "crosstab", original_crosstab)

    monkeypatch.setattr(
        heatmap_mod,
        "fisher_exact",
        lambda table: (_ for _ in ()).throw(ValueError("bad")),
    )
    with pytest.raises(ValueError, match="Fisher's exact test failed"):
        cns.confusionplot(
            confusion_df,
            x="pred",
            y="truth",
            add_pvalue=True,
            x_order=["neg", "pos"],
            y_order=["neg", "pos"],
        )

    bad_survival = survival_df.copy()
    bad_survival["time"] = bad_survival["time"] * 20
    cns.figure(120, 120)
    ax = cns.survivalplot(bad_survival, "time", "event", "group")
    assert ax.get_xlabel() == "Time (Months)"

    import lifelines

    original_logrank_test = lifelines.statistics.multivariate_logrank_test

    monkeypatch.setattr(
        lifelines.statistics,
        "multivariate_logrank_test",
        lambda *args, **kwargs: (_ for _ in ()).throw(RuntimeError("bad")),
    )
    with pytest.raises(RuntimeError, match="Log-rank test failed"):
        cns.survivalplot(survival_df, "time", "event", "group")
    monkeypatch.setattr(
        lifelines.statistics,
        "multivariate_logrank_test",
        original_logrank_test,
    )

    class BadCox:
        def fit(self, *args: object, **kwargs: object) -> None:
            raise RuntimeError("bad")

    monkeypatch.setattr(lifelines, "CoxPHFitter", BadCox)
    with pytest.raises(RuntimeError, match="Cox proportional hazards model failed"):
        cns.survivalplot(
            pd.DataFrame(
                {
                    "time": [1, 2, 3, 4, 5, 6],
                    "event": [1, 0, 1, 0, 1, 0],
                    "group": ["A", "A", "B", "B", "C", "C"],
                }
            ),
            "time",
            "event",
            "group",
        )
    with pytest.raises(RuntimeError, match="Could not compute hazard ratios"):
        cns.survivalplot(survival_df, "time", "event", "group")

    cns.figure(120, 120)
    cns.phyloplot(phylo_adata)

    fake_venn_obj = types.SimpleNamespace(
        get_label_by_id=lambda area: (_ for _ in ()).throw(AttributeError())
        if area in {"100", "110"}
        else types.SimpleNamespace(set_fontsize=lambda size: None),
        get_patch_by_id=lambda area: (_ for _ in ()).throw(AttributeError())
        if area in {"100", "110"}
        else types.SimpleNamespace(
            set_edgecolor=lambda color: None, set_linewidth=lambda width: None
        ),
    )
    monkeypatch.setitem(
        sys.modules,
        "matplotlib_venn",
        types.SimpleNamespace(
            venn2=lambda *args, **kwargs: fake_venn_obj,
            venn3=lambda *args, **kwargs: fake_venn_obj,
        ),
    )
    cns.figure(120, 120)
    assert (
        cns.vennplot([{1, 2}, {2, 3}, {3, 4}], labels=["A", "B", "C"]) is fake_venn_obj
    )


def test_remaining_visual_gap_coverage(
    confusion_df: pd.DataFrame,
    heatmap_adata: ad.AnnData,
    numeric_df: pd.DataFrame,
    volcano_df: pd.DataFrame,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    real_isinstance = builtins.isinstance

    def fake_isinstance(obj: object, typ: Any) -> bool:
        if typ is object and getattr(obj, "kind", None) in {"b", "i", "u", "f"}:
            return False
        return real_isinstance(obj, typ)

    monkeypatch.setattr(heatmap_mod, "isinstance", fake_isinstance, raising=False)
    cns.figure(180, 180)
    cmp = cns.heatmapplot(
        heatmap_adata,
        row_annotation=["score"],
        col_annotation=["importance"],
        cmap="parula",
    )
    assert cmp.ax_heatmap is not None

    monkeypatch.setattr(
        heatmap_mod.pd,
        "crosstab",
        lambda *args, **kwargs: pd.DataFrame(
            [[np.nan, 1], [1, 1]],
            index=pd.Index(["neg", "pos"]),
            columns=pd.Index(["neg", "pos"]),
        ),
    )
    with pytest.raises(ValueError, match="contains NaN at position \\[0,0\\]"):
        cns.confusionplot(
            confusion_df,
            x="pred",
            y="truth",
            x_order=["neg", "pos"],
            y_order=["neg", "pos"],
            annot=True,
        )

    monkeypatch.setattr(
        heatmap_mod.pd,
        "Categorical",
        lambda values, categories, ordered: values,
    )
    monkeypatch.setattr(
        heatmap_mod.pd,
        "crosstab",
        lambda *args, **kwargs: pd.DataFrame(
            [[1, 0], [0, 1]],
            index=pd.Index(["neg", "pos"]),
            columns=pd.Index(["pos1", "pos2"]),
        ),
    )
    with pytest.raises(ValueError, match="Could not find negative label in x_order"):
        cns.confusionplot(
            confusion_df,
            x="pred",
            y="truth",
            x_order=["pos", "pos"],
            y_order=["neg", "pos"],
            positive_x="pos",
            add_pvalue=True,
            annot=False,
        )

    class MarkerHandle:
        def __init__(self) -> None:
            self.marker_size = None

        def set_markersize(self, size: float) -> None:
            self.marker_size = size

    reg_legend = types.SimpleNamespace(legend_handles=[MarkerHandle()])
    monkeypatch.setattr(
        type(plt.gca()),
        "get_legend",
        lambda self: reg_legend,
        raising=False,
    )
    cns.figure(120, 120)
    cns.regplot(numeric_df, x="x", y="y", color="color_group", s=9)
    assert reg_legend.legend_handles[0].marker_size is not None

    class SizeHandle:
        def __init__(self) -> None:
            self.sizes = None

        def set_sizes(self, sizes: list[float]) -> None:
            self.sizes = sizes

    volcano_legend = types.SimpleNamespace(legend_handles=[SizeHandle()])
    monkeypatch.setattr(
        type(plt.gca()),
        "get_legend",
        lambda self: volcano_legend,
        raising=False,
    )
    monkeypatch.setattr(genomics_mod.cns, "take_legend_out", lambda: None)
    cns.figure(120, 120)
    cns.volcanoplot(volcano_df)
    assert volcano_legend.legend_handles[0].sizes == [20]

    cox_like_model = types.SimpleNamespace(
        name="cox",
        hue="group",
        results=pd.DataFrame(
            {
                "display_label": ["Age", "Stage"],
                "exp(coef)": [1.2, 0.9],
                "log10_pvalue": [1.1, 0.7],
                "exp(coef) lower_err": [0.1, 0.1],
                "exp(coef) upper_err": [0.2, 0.15],
                "hue_group": ["All", "All"],
            }
        ),
    )
    cns.figure(120, 120)
    ax = cns.forestplot(cox_like_model)
    assert ax.get_xlabel() == "Hazard ratio (95% CI)"

    monkeypatch.setattr(
        specialized_mod, "validate_dataframe", lambda data, name, fn: None
    )
    with pytest.raises(TypeError, match="Internal type validation failed"):
        cns.forestplot(
            types.SimpleNamespace(name="cox", results=["not", "a", "dataframe"])
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    pass

from collections.abc import Mapping, Set as AbstractSet

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

import cnsplots as cns


def _normalize_text_position(text: Any) -> None:
    """Coerce singleton array coordinates from upsetplot into scalar positions."""

    def _to_scalar(value: Any) -> Any:
        if isinstance(value, np.ndarray) and value.size == 1:
            return float(value.reshape(-1)[0])
        return value

    x, y = text.get_position()
    text.set_position((_to_scalar(x), _to_scalar(y)))


def upsetplot(sets: Mapping[str, AbstractSet[Any] | list[Any]], **kwargs: Any) -> dict:
    """
    Create an UpSet plot for visualizing set intersections.

    This function creates an UpSet plot, an advanced alternative to Venn diagrams
    for visualizing intersections among multiple sets.

    Parameters
    ----------
    sets : dict
        Dictionary mapping set names (str) to sets or array-like collections.
    **kwargs
        Additional keyword arguments passed to `upsetplot.UpSet`.

    Returns
    -------
    dict
        Dictionary of matplotlib Axes objects from the UpSet plot.

    See Also
    --------
    vennplot : Create a Venn diagram for 2-3 sets.

    Examples
    --------
    >>> import cnsplots as cns
    >>> sets = {
    ...     "Set A": [1, 2, 3, 4, 5],
    ...     "Set B": [3, 4, 5, 6, 7],
    ...     "Set C": [5, 6, 7, 8, 9],
    ... }
    >>> axes = cns.upsetplot(sets)
    >>> axes["intersections"].set_ylabel("Intersection Size")

    >>> # Limit to larger intersections
    >>> axes = cns.upsetplot(sets, min_subset_size=10)
    """
    # Validate inputs
    if not isinstance(sets, dict):
        raise TypeError(
            f"[upsetplot] Parameter 'sets' must be a dictionary, got {type(sets).__name__}"
        )
    if not sets:
        raise ValueError("[upsetplot] Parameter 'sets' cannot be empty")

    import upsetplot as usp

    normalized_sets: dict[str, set] = {
        k: (v if isinstance(v, set) else set(v)) for k, v in sets.items()
    }
    memberships = []
    all_items: set = set().union(*normalized_sets.values())
    for item in all_items:
        membership = [name for name, s in normalized_sets.items() if item in s]
        memberships.append(membership)
    data = usp.from_memberships(memberships)
    # Set default subset_size to "count" to handle non-unique groups
    kwargs.setdefault("subset_size", "count")
    upset = usp.UpSet(data, element_size=17, show_counts="{:,}", **kwargs)
    axes = upset.plot()
    plt.grid(False)
    ax_tot = axes.get("totals")
    cns.setup_ax(axes["matrix"])
    cns.setup_ax(axes["shading"])
    cns.setup_ax(axes["intersections"])
    axes["matrix"].tick_params(axis="both", which="both", length=0)
    for txt in axes["intersections"].texts:
        _normalize_text_position(txt)
        txt.set_size(cns.settings.fontsize_legend)
    if ax_tot is not None:
        cns.setup_ax(ax_tot)
        for txt in ax_tot.texts:
            _normalize_text_position(txt)
            txt.set_size(cns.settings.fontsize_legend)
        pos_mat = ax_tot.get_position()
        dx = 0.03
        new_pos = [pos_mat.x0 + dx, pos_mat.y0, pos_mat.width, pos_mat.height]
        ax_tot.set_position(new_pos)

    return axes


def vennplot(lists: list[set], labels: tuple[str, ...] | list[str]) -> Any:
    """
    Create a Venn diagram for 2 or 3 sets.

    This function generates a Venn diagram showing overlaps between 2 or 3 sets
    using colored, semi-transparent circles with intersection counts.

    Parameters
    ----------
    lists : list of set or array-like
        List of 2 or 3 sets or array-like collections to compare.
    labels : tuple or list of str
        Labels for each set, in the same order as lists.

    Returns
    -------
    matplotlib_venn.VennDiagram
        The Venn diagram object containing the plot elements.

    See Also
    --------
    upsetplot : Create an UpSet plot for multiple set intersections.

    Examples
    --------
    >>> import cnsplots as cns
    >>> set1 = {1, 2, 3, 4, 5}
    >>> set2 = {3, 4, 5, 6, 7}
    >>> set3 = {5, 6, 7, 8, 9}
    >>> venn = cns.vennplot([set1, set2], labels=["Group A", "Group B"])

    >>> # Three-way Venn diagram
    >>> venn = cns.vennplot(
    ...     [set1, set2, set3], labels=["Treatment A", "Treatment B", "Treatment C"]
    ... )
    """
    # Validate inputs
    if not isinstance(lists, list):
        raise TypeError(
            f"[vennplot] Parameter 'lists' must be a list, got {type(lists).__name__}"
        )
    if len(lists) not in [2, 3]:
        raise ValueError(
            f"[vennplot] Parameter 'lists' must contain 2 or 3 sets, got {len(lists)}"
        )
    if len(labels) != len(lists):
        raise ValueError(
            f"[vennplot] Length of 'labels' ({len(labels)}) must match length of 'lists' ({len(lists)})"
        )

    import matplotlib_venn as venn

    lists = [s if isinstance(s, AbstractSet) else set(s) for s in lists]
    func: Any
    if len(lists) == 2:
        areas = ["10", "01", "11"]
        func = venn.venn2
        names = ["A", "B"]
        colors = sns.color_palette(n_colors=2)
    else:
        areas = ["100", "010", "001", "110", "101", "011", "111"]
        func = venn.venn3
        names = ["A", "B", "C"]
        colors = sns.color_palette(n_colors=3)
    ax = func(lists, tuple(labels), set_colors=colors, alpha=0.8)  # type: ignore[arg-type]
    for area in areas:
        try:
            ax.get_label_by_id(area).set_fontsize(6)
            ax.get_patch_by_id(area).set_edgecolor("black")
            ax.get_patch_by_id(area).set_linewidth(0.5)
        except AttributeError:
            pass
    for area in names:
        ax.get_label_by_id(area).set_fontsize(7)

    return ax



"""Plot functions for cnsplots."""

from cnsplots.plots._categorical import (
    barplot,
    donutplot,
    lollipopplot,
    pieplot,
    stackplot,
    stripplot,
)
from cnsplots.plots._distribution import (
    boxplot,
    distplot,
    histplot,
    kdeplot,
    qqplot,
    ridgeplot,
    violinplot,
)
from cnsplots.plots._genomics import gseaplot, volcanoplot
from cnsplots.plots._heatmap import confusionplot, dotplot, heatmapplot
from cnsplots.plots._regression import lineplot, regplot, scatterplot, slopeplot
from cnsplots.plots._sets import upsetplot, vennplot
from cnsplots.plots._specialized import forestplot, phyloplot, rocplot, sankeyplot
from cnsplots.plots._survival import cumulativeincidenceplot, survivalplot

__all__ = [
    "barplot",
    "boxplot",
    "confusionplot",
    "cumulativeincidenceplot",
    "distplot",
    "donutplot",
    "dotplot",
    "forestplot",
    "gseaplot",
    "heatmapplot",
    "histplot",
    "kdeplot",
    "lineplot",
    "lollipopplot",
    "phyloplot",
    "pieplot",
    "qqplot",
    "regplot",
    "ridgeplot",
    "rocplot",
    "sankeyplot",
    "scatterplot",
    "slopeplot",
    "stackplot",
    "stripplot",
    "survivalplot",
    "upsetplot",
    "vennplot",
    "violinplot",
    "volcanoplot",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import pandas as pd
    from matplotlib.axes import Axes

import matplotlib.pyplot as plt
import num2tex
import numpy as np
import palettable.colorbrewer.qualitative  # noqa: F401  # ensure submodule is importable
import scipy as sp
import seaborn as sns

from cnsplots._validation import (
    validate_column_type,
    validate_columns_exist,
    validate_dataframe,
    validate_dataframe_not_empty,
)


def regplot(
    data: pd.DataFrame,
    x: str,
    y: str,
    hue: str | None = None,
    s: float = 3,
    color="black",
    **kwargs: Any,
) -> Axes:
    """
    Create a regression plot with linear fit and correlation statistics.

    This function creates a scatter plot with a fitted regression line and displays
    Pearson correlation coefficient and p-value.

    Parameters
    ----------
    data : pd.DataFrame
        The input DataFrame containing the data to be plotted.
    x : str
        Column name for the independent variable (x-axis).
    y : str
        Column name for the dependent variable (y-axis).
    hue : str, optional
        Column name for grouping. When set, separate regression lines and
        correlation statistics are drawn for each group.
    s : float, default: 3
        Size of scatter plot markers.
    color : str, default: "black"
        Either a matplotlib color string (e.g. ``"black"``, ``"#ff0000"``) or the
        name of a column in *data*.  When a column name is given, the scatter
        points are colored by the unique values of that column and a legend is
        added, while a single overall regression line and correlation statistic
        are shown.  If *hue* is also specified, *hue* takes precedence and
        *color* is ignored.
    **kwargs
        Additional keyword arguments passed to `seaborn.regplot`.

    Returns
    -------
    matplotlib.axes.Axes
        The matplotlib Axes object containing the plot.

    See Also
    --------
    scatterplot : Create a scatter plot without regression line.
    kdeplot : Create a kernel density plot for distributions.

    Examples
    --------
    >>> import cnsplots as cns
    >>> ax = cns.regplot(data=df, x="age", y="expression")
    >>> ax.set_title("Age vs Expression")

    >>> # Grouped regression with separate fits
    >>> ax = cns.regplot(data=df, x="dose", y="response", hue="cell_line")
    >>> ax.set_xlabel("Drug Dose")

    >>> # Color points by a column (single regression line)
    >>> ax = cns.regplot(data=df, x="age", y="expression", color="cell_type")
    """
    # Validate inputs
    validate_dataframe(data, "data", "regplot")
    columns_to_check = [x, y]
    if hue is not None:
        columns_to_check.append(hue)
    validate_columns_exist(data, columns_to_check, "regplot")
    validate_dataframe_not_empty(data, "regplot")

    # Validate numeric columns
    validate_column_type(data, x, ["numeric"], "regplot")
    validate_column_type(data, y, ["numeric"], "regplot")

    ax = plt.gca()
    args = {
        "line_kws": {"lw": 1.2},
        "scatter_kws": {"s": s, "alpha": 1, "edgecolor": None},
    }
    args.update(kwargs)
    palette = plt.rcParams["axes.prop_cycle"].by_key()["color"]
    color_is_column = isinstance(color, str) and color in data.columns
    if hue:
        for idx, hue_val in enumerate(data[hue].unique()):
            subset = data[data[hue] == hue_val]
            ax = sns.regplot(
                data=subset,
                x=x,
                y=y,
                ax=ax,
                color=palette[idx],
                label=hue_val,
                **args,
            )
            rho, p_value = sp.stats.pearsonr(subset[x], subset[y])
            ax.text(
                0.05,
                0.95 - 0.08 * idx,
                rf"{hue_val}: $\rho$={rho:.2f},"
                rf" P=${num2tex.num2tex(p_value, precision=2):.2g}$",
                color=palette[idx],
                transform=ax.transAxes,
                ha="left",
                va="top",
            )
        ax.legend(title=hue)
    elif color_is_column:
        line_args = {k: v for k, v in args.items() if k != "scatter_kws"}
        ax = sns.regplot(
            data=data,
            x=x,
            y=y,
            ax=ax,
            color="black",
            scatter=False,
            **line_args,
        )
        unique_vals = data[color].unique()
        for idx, val in enumerate(unique_vals):
            subset = data[data[color] == val]
            ax.scatter(
                subset[x],
                subset[y],
                s=s,
                color=palette[idx % len(palette)],
                label=val,
                alpha=1,
                edgecolors="none",
            )
        rho, p_value = sp.stats.pearsonr(data[x], data[y])
        ax.text(
            0.05,
            0.95,
            rf"$\rho$={rho:.2f}, $P={num2tex.num2tex(p_value, precision=2):.2g}$",
            color="black",
            transform=ax.transAxes,
            ha="left",
            va="top",
        )
        ax.legend(title=color)
        for handle in ax.get_legend().legend_handles:
            if hasattr(handle, "set_sizes"):
                handle.set_sizes([2 * s])
            elif hasattr(handle, "set_markersize"):
                handle.set_markersize(2 * np.sqrt(s / np.pi))
    else:
        ax = sns.regplot(
            data=data,
            x=x,
            y=y,
            ax=ax,
            color=color,
            **args,
        )
        rho, p_value = sp.stats.pearsonr(data[x], data[y])
        ax.text(
            0.05,
            0.95,
            rf"$\rho$={rho:.2f}, $P={num2tex.num2tex(p_value, precision=2):.2g}$",
            color=color,
            transform=ax.transAxes,
            ha="left",
            va="top",
        )

    return ax


def scatterplot(
    data: pd.DataFrame, x: str, y: str, s: float = 7, **kwargs: Any
) -> Axes:
    """
    Create a scatter plot with automatic legend size correction.

    Parameters
    ----------
    data : pd.DataFrame
        The input DataFrame containing the data to be plotted.
    x : str
        Column name for the x-axis variable.
    y : str
        Column name for the y-axis variable.
    s : float, default: 7
        Size of scatter plot markers.
    **kwargs
        Additional keyword arguments passed to `seaborn.scatterplot`.

    Returns
    -------
    matplotlib.axes.Axes
        The matplotlib Axes object containing the plot.

    See Also
    --------
    regplot : Create a scatter plot with regression line.
    stripplot : Create a categorical scatter plot.
    volcanoplot : Create a volcano plot for differential expression.

    Examples
    --------
    >>> import cnsplots as cns
    >>> ax = cns.scatterplot(data=df, x="PC1", y="PC2", s=10)
    >>> ax.set_title("PCA Plot")

    >>> # With grouping by category
    >>> ax = cns.scatterplot(
    ...     data=df, x="UMAP1", y="UMAP2", hue="cell_type", s=5, alpha=0.7
    ... )
    >>> ax.set_xlabel("UMAP Dimension 1")
    """
    # Validate inputs
    validate_dataframe(data, "data", "scatterplot")
    validate_columns_exist(data, [x, y], "scatterplot")
    validate_dataframe_not_empty(data, "scatterplot")

    ax = sns.scatterplot(data=data, x=x, y=y, s=s, edgecolor=None, **kwargs)

    if ax.get_legend() is not None:
        for handle in ax.get_legend().legend_handles:
            if hasattr(handle, "set_sizes"):
                handle.set_sizes([s])
            elif hasattr(handle, "set_markersize"):
                handle.set_markersize(2 * np.sqrt(s / np.pi))

    return ax


def lineplot(**kwargs: Any) -> Axes:
    """
    Create a line plot (wrapper around seaborn.lineplot).

    Parameters
    ----------
    **kwargs
        Keyword arguments passed directly to `seaborn.lineplot`.

    Returns
    -------
    matplotlib.axes.Axes
        The matplotlib Axes object containing the plot.

    See Also
    --------
    regplot : Create a regression plot with linear fit.
    scatterplot : Create a scatter plot without connecting lines.
    survivalplot : Create a Kaplan-Meier survival plot.

    Examples
    --------
    >>> import cnsplots as cns
    >>> ax = cns.lineplot(data=df, x="time", y="value")
    >>> ax.set_title("Time Series")

    >>> # Multiple groups with error bands
    >>> ax = cns.lineplot(
    ...     data=df, x="timepoint", y="expression", hue="treatment", errorbar="se"
    ... )
    >>> ax.set_ylabel("Gene Expression")
    """
    # Validate inputs if provided in kwargs
    if "data" in kwargs:
        validate_dataframe(kwargs["data"], "data", "lineplot")
        data = kwargs["data"]
        columns_to_check = []
        if "x" in kwargs:
            columns_to_check.append(kwargs["x"])
        if "y" in kwargs:
            columns_to_check.append(kwargs["y"])
        if columns_to_check:
            validate_columns_exist(data, columns_to_check, "lineplot")

    ax = sns.lineplot(**kwargs)
    return ax


def slopeplot(data: pd.DataFrame, x: str, y: str, hue: str) -> Axes:
    """
    Create a slope plot showing paired changes between two conditions.

    Parameters
    ----------
    data : pd.DataFrame
        The input DataFrame containing paired observations.
    x : str
        Column name for the categorical variable defining the two conditions.
    y : str
        Column name for the continuous variable to plot.
    hue : str
        Column name for the grouping variable. Must have exactly two unique values.

    Returns
    -------
    matplotlib.axes.Axes
        The matplotlib Axes object containing the plot.

    See Also
    --------
    lineplot : Create a line plot for time series.
    scatterplot : Create a scatter plot.
    regplot : Create a regression plot.

    Examples
    --------
    >>> import cnsplots as cns
    >>> # Compare before/after measurements for multiple patients
    >>> ax = cns.slopeplot(
    ...     data=df,
    ...     x="patient_id",
    ...     y="tumor_size",
    ...     hue="timepoint",
    ... )
    >>> ax.set_ylabel("Tumor Size (mm)")

    >>> # Compare two treatments across sites
    >>> ax = cns.slopeplot(data=df, x="site", y="response_rate", hue="treatment")
    """
    # Validate inputs
    validate_dataframe(data, "data", "slopeplot")
    validate_columns_exist(data, [x, y, hue], "slopeplot")
    validate_dataframe_not_empty(data, "slopeplot")

    # https://cduvallet.github.io/posts/2018/03/slopegraphs-in-python
    red = palettable.colorbrewer.qualitative.Set1_9.hex_colors[0]
    blue = palettable.colorbrewer.qualitative.Set1_9.hex_colors[1]
    hues = data[hue].unique()

    ax = plt.gca()

    sites = []
    i = 1.0
    for site, subdf in data.groupby(x):
        sites.append(site)
        h = subdf[subdf[hue] == hues[0]][y].values
        d = subdf[subdf[hue] == hues[1]][y].values

        x1 = i - 0.2
        x2 = i + 0.2

        line_colors = (h - d) > 0
        line_colors = [blue if j else red for j in line_colors]

        alphas = [0.4] * len(line_colors)

        for hi, di, ci, ai in zip(h, d, line_colors, alphas):
            ax.plot([x1, x2], [hi, di], c=ci, alpha=ai)

        ax.scatter(len(h) * [x1], h, c=blue, s=10, lw=0.5, label=hues[0])
        ax.scatter(len(d) * [x2], d, c=red, s=10, lw=0.5, label=hues[1])

        i += 1

    ax.set_xticks(list(1 + np.arange(len(sites))))
    ax.set_ylabel(y)
    _ = ax.set_xticklabels(sites)

    handles, labels = ax.get_legend_handles_labels()
    lgd = ax.legend(
        handles[0:2],
        labels[0:2],
        loc="upper center",
        bbox_to_anchor=(0.5, 1.15),
        ncol=2,
    )
    for handle in lgd.legend_handles:
        if hasattr(handle, "set_sizes"):
            handle.set_sizes([12])  # type: ignore[union-attr]

    return ax



from __future__ import annotations

from pathlib import Path
from tempfile import TemporaryDirectory
import warnings
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from lxml.etree import _Element

import subprocess

import matplotlib.pyplot as plt
from lxml import etree
from matplotlib.text import Text


def _collect_bold_texts() -> set[str]:
    """Collect text strings that are bold in the current matplotlib figure."""
    bold_texts: set[str] = set()
    fig = plt.gcf()
    for ax in fig.get_axes():
        for artist in ax.get_children():
            if not isinstance(artist, Text):
                continue
            txt = artist.get_text().strip()
            if not txt:
                continue
            weight = artist.get_fontweight()
            if weight in ("bold", "heavy", "extra bold", "semibold", "demibold") or (
                isinstance(weight, (int, float)) and weight >= 600
            ):
                bold_texts.add(txt)
                for line in txt.split("\n"):
                    line = line.strip()
                    if line:
                        bold_texts.add(line)
    return bold_texts


def _save_plain_svg(filepath: str, message: str) -> None:
    """Save a standard matplotlib SVG and surface why the optimized path was skipped."""
    warnings.warn(message, RuntimeWarning, stacklevel=2)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        plt.savefig(filepath, format="svg")


def _save_svg(filepath: str, root: str) -> None:
    bold_texts = _collect_bold_texts()
    stem = Path(root).name or Path(filepath).stem or "cnsplots"

    with TemporaryDirectory(prefix=f"{stem}-svg-") as tmp_dir:
        tmp_dir_path = Path(tmp_dir)
        tmp_pdf = tmp_dir_path / f"{stem}.pdf"
        tmp_svg = tmp_dir_path / "1.svg"
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            plt.savefig(tmp_pdf)
        try:
            subprocess.run(
                [
                    "mutool",
                    "convert",
                    "-F",
                    "svg",
                    "-O",
                    "text=text",
                    "-o",
                    str(tmp_dir_path / "%d.svg"),
                    str(tmp_pdf),
                ],
                check=True,
                capture_output=True,
            )
            _correct_svg(str(tmp_svg), filepath, bold_texts)
        except FileNotFoundError:
            _save_plain_svg(
                filepath,
                "MuPDF's `mutool` is unavailable; saved a standard matplotlib SVG instead.",
            )
        except subprocess.CalledProcessError as exc:
            stderr = (exc.stderr or b"").decode(errors="replace").strip()
            detail = f" ({stderr})" if stderr else ""
            _save_plain_svg(
                filepath,
                "MuPDF's `mutool` failed during SVG conversion"
                f"{detail}; saved a standard matplotlib SVG instead.",
            )


def _correct_svg(
    input_file: str, output_file: str, bold_texts: set[str] | None = None
) -> None:
    """
    Process an SVG file to ungroup both text elements and clipped elements.

    Args:
        input_file (str): Path to the input SVG file
        output_file (str): Path to save the processed SVG file
        bold_texts (set[str] | None): Text strings that should have font-weight bold
    """
    # Read the SVG file
    with open(input_file) as f:
        svg_content = f.read()

    # NOTE: clip-path attributes and clipPath definitions are intentionally
    # preserved so that axes clipping (e.g. from set_xlim / set_ylim) is
    # rendered correctly.  The _flatten_groups helper propagates clip-path
    # from parent <g> elements to their children when groups are dissolved.

    # Parse the modified SVG with lxml
    parser = etree.XMLParser(remove_blank_text=True)
    root = etree.fromstring(svg_content.encode("utf-8"), parser)

    # Define SVG namespace
    ns = {"svg": "http://www.w3.org/2000/svg"}

    # Process text elements
    _process_text_elements_lxml(root, ns)

    # Restore bold font-weight lost during PDF→SVG conversion
    _restore_bold_fonts(root, ns, bold_texts or set())

    # Flatten any remaining g elements
    _flatten_groups(root, ns)

    # Create an ElementTree from the root element
    tree = etree.ElementTree(root)

    # Write the modified SVG to output file
    tree.write(output_file, encoding="utf-8", xml_declaration=True, pretty_print=True)


def _process_text_elements_lxml(root: _Element, ns: dict[str, str]) -> None:
    """Process text elements to ungroup them using lxml."""
    # Find all text elements with tspan children
    text_elements = root.xpath(".//svg:text[svg:tspan]", namespaces=ns)

    for text in text_elements:
        # Find parent of text element
        parent = text.getparent()
        if parent is None:
            continue

        # Find all tspan elements within this text
        tspans = text.xpath("./svg:tspan", namespaces=ns)

        for tspan in tspans:
            # Create a new text element
            new_text = etree.Element("{http://www.w3.org/2000/svg}text")

            # Copy attributes from original text element
            for attr, value in text.attrib.items():
                new_text.set(attr, value)

            # Copy content and attributes from tspan
            new_text.text = tspan.text
            for attr, value in tspan.attrib.items():
                new_text.set(attr, value)

            # Add the new text element before the original
            parent.insert(parent.index(text), new_text)

        # Remove the original text element
        parent.remove(text)


def _restore_bold_fonts(
    root: _Element, ns: dict[str, str], bold_texts: set[str]
) -> None:
    """Restore font-weight="bold" on text elements whose content was bold in matplotlib.

    Mutool drops bold information during PDF→SVG conversion, so we match text
    content against the set collected from matplotlib before saving.
    """
    if not bold_texts:
        return

    for text_el in root.xpath(".//svg:text", namespaces=ns):
        content = (text_el.text or "").strip()
        if content in bold_texts:
            text_el.set("font-weight", "bold")


def _flatten_groups(root: _Element, ns: dict[str, str]) -> None:
    """Flatten all group elements by moving their children to parent.

    When a ``<g>`` carries a ``clip-path`` attribute the attribute is
    propagated to each child element so that clipping (e.g. from
    ``set_xlim`` / ``set_ylim``) is preserved after the group is removed.
    """
    # Iteratively flatten groups until no more flattening occurs
    while True:
        # Find g elements
        g_elements = root.xpath("//svg:g", namespaces=ns)

        if not g_elements:
            break

        flattened = False
        for g in g_elements:
            parent = g.getparent()
            if parent is None:
                continue

            clip_path = g.get("clip-path")

            # Get index of g in parent
            g_index = parent.index(g)

            # Move all children of g to parent
            children = list(g)
            for child in children:
                # Propagate clip-path from the group to children that
                # don't already have their own clip-path.
                if clip_path is not None and child.get("clip-path") is None:
                    child.set("clip-path", clip_path)
                g.remove(child)
                parent.insert(g_index, child)
                g_index += 1

            # Remove empty g element
            parent.remove(g)
            flattened = True
            break  # Break after flattening one g to avoid modifying during iteration

        if not flattened:
            break  # Exit if no more g elements were flattened



from __future__ import annotations

import sys
import types

import anndata as ad
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest
from matplotlib.axes import Axes

import cnsplots as cns


def test_survival_plots(
    survival_df: pd.DataFrame,
    survival_three_group_df: pd.DataFrame,
    competing_risk_df: pd.DataFrame,
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
) -> None:
    cns.figure(120, 120)
    ax = cns.survivalplot(
        survival_df, "time", "event", "group", hue_order=["Treatment", "Control"]
    )
    assert ax.get_ylabel() == "Overall survival probability"
    assert "HR =" in ax.texts[0].get_text()
    assert "multivariate log-rank test" in capsys.readouterr().out

    cns.figure(120, 120)
    ax2 = cns.survivalplot(survival_three_group_df, "time", "event", "group")
    assert ax2.get_xlabel() == "Time (Years)"
    assert "trend" in capsys.readouterr().out

    added: dict[str, object] = {}
    import lifelines

    monkeypatch.setattr(
        lifelines.plotting,
        "add_at_risk_counts",
        lambda *fitters, **kwargs: added.update({"fitters": fitters, **kwargs}),
    )
    cns.figure(120, 120)
    ax3 = cns.cumulativeincidenceplot(
        competing_risk_df,
        "time",
        "event",
        "group",
        show_risk_table=True,
        xticks=[0, 2, 4, 6, 8],
    )
    assert list(ax3.get_xticks()) == [0, 2, 4, 6, 8]
    assert added["rows_to_show"] == ["At risk"]
    assert "Gray's test" in capsys.readouterr().out

    cns.figure(120, 120)
    single_group = competing_risk_df[competing_risk_df["group"] == "A"].copy()
    ax4 = cns.cumulativeincidenceplot(single_group, "time", "event", "group")
    assert ax4 is plt.gca()


def test_confusionplot_metrics_and_errors(confusion_df: pd.DataFrame) -> None:
    cns.figure(120, 120)
    ax = cns.confusionplot(
        confusion_df,
        x="pred",
        y="truth",
        add_pvalue=True,
        x_order=["neg", "pos"],
        y_order=["neg", "pos"],
        positive_x="pos",
        positive_y="pos",
    )
    assert ax.get_xlabel() == "pred"
    assert len(plt.gcf().axes) == 2

    cns.figure(120, 120)
    ax2 = cns.confusionplot(confusion_df, x="pred", y="truth", annot=False)
    assert ax2 is plt.gca()

    with pytest.raises(ValueError, match="2x2 confusion matrix"):
        cns.confusionplot(
            pd.DataFrame({"pred": ["a", "b", "c"], "truth": ["a", "b", "c"]}),
            x="pred",
            y="truth",
            add_pvalue=True,
        )

    with pytest.raises(ValueError, match="Could not find negative label in y_order"):
        cns.confusionplot(
            confusion_df,
            x="pred",
            y="truth",
            add_pvalue=True,
            x_order=["neg", "pos"],
            y_order=["pos"],
            positive_y="pos",
        )

    with pytest.raises(ValueError, match="2x2 confusion matrix"):
        cns.confusionplot(
            confusion_df,
            x="pred",
            y="truth",
            add_pvalue=True,
            x_order=["pos"],
            y_order=["neg", "pos"],
            positive_x="pos",
        )

    with pytest.raises(ValueError, match="Categorical categories cannot be null"):
        cns.confusionplot(
            pd.DataFrame({"pred": ["neg", "pos"], "truth": ["neg", np.nan]}),
            x="pred",
            y="truth",
        )


def test_heatmap_and_dotplot(
    heatmap_adata: ad.AnnData,
    dotplot_df: pd.DataFrame,
) -> None:
    cns.figure(180, 180)
    cmp = cns.heatmapplot(
        heatmap_adata,
        layer="scaled",
        row_annotation=["cluster", "score"],
        col_annotation=["pathway", "importance"],
        row_split="cluster",
        col_split="pathway",
        row_cluster=True,
        col_cluster=True,
        colors={"cluster": {"A": "#111111", "B": "#222222"}},
        cmap="parula",
    )
    assert cmp.ax_heatmap is not None

    cns.figure(180, 180)
    cmp2 = cns.heatmapplot(
        heatmap_adata,
        row_annotation=["cluster"],
        col_annotation=["pathway"],
        colors={"cluster": {"missing": "#111111"}},
        cmap="Set1",
    )
    assert cmp2.ax_heatmap is not None

    cns.figure(160, 160)
    dp = cns.dotplot(
        dotplot_df,
        x="sample",
        y="gene",
        color="mean_expr",
        size="pct_expr",
        value="score",
    )
    assert dp.ax_heatmap is not None

    cns.figure(160, 160)
    with pytest.raises(ValueError, match="Length mismatch"):
        cns.dotplot(
            dotplot_df[["sample", "gene", "mean_expr", "pct_expr"]],
            x="sample",
            y="gene",
            color="mean_expr",
            size="pct_expr",
        )


def test_genomics_plots(
    volcano_df: pd.DataFrame,
    gsea_plot_df: pd.DataFrame,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    fake_adjust = types.SimpleNamespace(adjust_text=lambda *args, **kwargs: None)
    monkeypatch.setitem(sys.modules, "adjustText", fake_adjust)

    cns.figure(120, 120)
    ax = cns.volcanoplot(volcano_df)
    assert ax.get_xlabel() == "log2(fold change)"

    cns.figure(120, 120)
    ax2 = cns.volcanoplot(volcano_df, show_list=["GENE1", "GENE6"])
    assert ax2.get_ylabel() == "–log10(adjusted p-value)"

    def fake_dotplot(
        data: pd.DataFrame,
        cmap: str,
        y: str,
        x: str,
        cutoff: float,
        column: str,
        ax: Axes,
        top_term: int,
        size: float,
    ) -> None:
        scatter = ax.scatter(data[x], np.arange(len(data)), c=data[column], s=20)
        fig = plt.gcf()
        cbar = fig.colorbar(scatter, ax=ax)
        cbar.set_label(column)
        handles = [plt.Line2D([], [], marker="o", linestyle="none", color="black")]
        ax.legend(handles, ["20"], title="size")

    monkeypatch.setitem(
        sys.modules, "gseapy", types.SimpleNamespace(dotplot=fake_dotplot)
    )
    cns.figure(160, 140)
    ax3 = cns.gseaplot(gsea_plot_df, y="Clean_Term", color="NES", top_term=2)
    assert ax3.get_xlabel() == "Normalized Enrichment Score (NES)"


def test_sets_validation_errors(sets_fixture: dict[str, set[int]]) -> None:
    with pytest.raises(TypeError, match="must be a dictionary"):
        cns.upsetplot([])  # type: ignore[arg-type]
    with pytest.raises(ValueError, match="cannot be empty"):
        cns.upsetplot({})
    with pytest.raises(TypeError, match="must be a list"):
        cns.vennplot(tuple(sets_fixture.values()), labels=["A", "B"])  # type: ignore[arg-type]
    with pytest.raises(ValueError, match="must contain 2 or 3 sets"):
        cns.vennplot([set()], labels=["A"])
    with pytest.raises(ValueError, match="Length of 'labels'"):
        cns.vennplot([set(), set()], labels=["A"])



from __future__ import annotations


import matplotlib.pyplot as plt
import pandas as pd
import pytest

import cnsplots as cns


def test_boxplot_and_violinplot(
    categorical_df: pd.DataFrame,
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
) -> None:
    calls: list[object] = []
    monkeypatch.setattr(
        cns.utils,
        "_p_value_helper",
        lambda *args, **kwargs: calls.append((args, kwargs)),
    )

    cns.figure(120, 120)
    ax = cns.boxplot(
        categorical_df,
        x="group",
        y="value",
        pairs=[("A", "B")],
        addcount=True,
        showoutliers=True,
        whis=(0, 100),
    )
    assert ax.get_xticklabels()[0].get_text().startswith("A")
    assert "minimum and maximum values" in capsys.readouterr().out
    assert calls

    cns.figure(120, 120)
    ax2 = cns.violinplot(
        categorical_df,
        x="group",
        y="value",
        pairs=[("A", "B")],
        add_box=True,
        addcount=True,
        hue="hue",
        split=True,
        inner="quart",
    )
    assert len(ax2.collections) > 0

    cns.figure(120, 120)
    ax3 = cns.violinplot(categorical_df, x="group", y="value", add_box=False)
    assert ax3 is plt.gca()


def test_barplot_and_lollipopplot(
    categorical_df: pd.DataFrame,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    calls: list[object] = []
    monkeypatch.setattr(
        cns.utils,
        "_p_value_helper",
        lambda *args, **kwargs: calls.append((args, kwargs)),
    )

    cns.figure(120, 120)
    ax = cns.barplot(
        categorical_df,
        x="group",
        y="value",
        pairs=[("A", "B")],
        addtip=True,
        palette="palette_group",
    )
    assert ax.get_legend() is not None
    assert ax.get_legend().get_title().get_text() == "palette_group"

    cns.figure(120, 120)
    ax2 = cns.lollipopplot(
        categorical_df,
        x="group",
        y="value",
        hue="hue",
        pairs=[(("A", "H1"), ("A", "H2"))],
        addtip=True,
        errorbar="ci",
        palette="Set1",
    )
    assert ax2.get_legend() is not None

    cns.figure(120, 120)
    ax3 = cns.lollipopplot(
        categorical_df.rename(columns={"group": "cat", "value": "num"}),
        x="num",
        y="cat",
        color="black",
        errorbar="sd",
        addtip=True,
        estimator="median",
    )
    assert ax3.get_yticklabels()[0].get_text()

    cns.figure(120, 120)
    ax4 = cns.lollipopplot(
        categorical_df,
        x="group",
        y="value",
        palette="palette_group",
        errorbar="se",
    )
    assert ax4.get_legend() is not None
    assert calls


def test_stack_strip_pie_and_donut_plots(
    categorical_df: pd.DataFrame,
    stack_df: pd.DataFrame,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    calls: list[object] = []
    monkeypatch.setattr(
        cns.utils,
        "_p_value_helper",
        lambda *args, **kwargs: calls.append((args, kwargs)),
    )

    cns.figure(120, 120)
    ax = cns.stackplot(
        stack_df,
        x="treatment",
        y="response",
        normalize=True,
        addtip=True,
        pairs=[("A", "B")],
        bar_order=["A", "B", "C"],
        stack_order=["Yes", "No"],
    )
    assert ax.get_ylabel() == "Frequency"

    cns.figure(120, 120)
    ax2 = cns.stackplot(
        categorical_df.rename(columns={"group": "treatment", "binary": "response"}),
        x="treatment",
        y="response",
        horizontal=True,
        normalize=False,
        pairs=[("A", "B")],
        bar_order=["A", "B", "C"],
        stack_order=["No", "Yes"],
    )
    assert ax2.get_xlabel() == "Count"

    cns.figure(120, 120)
    ax3 = cns.stripplot(
        categorical_df,
        x="group",
        y="value",
        hue="hue",
        showmeans=True,
        addcount=True,
    )
    assert ax3.get_legend() is not None

    cns.figure(120, 120)
    ax4 = cns.pieplot(
        categorical_df, x="group", legend="left", hue_order=["C", "B", "A"]
    )
    assert ax4.get_legend() is not None

    cns.figure(120, 120)
    ax5 = cns.donutplot(
        categorical_df, x="group", legend="top", hue_order=["A", "B", "C"]
    )
    assert ax5.get_legend() is not None
    assert any(text.get_text() == "group" for text in ax5.texts)
    assert calls


def test_distribution_wrappers(
    numeric_df: pd.DataFrame,
    categorical_df: pd.DataFrame,
    capsys: pytest.CaptureFixture[str],
) -> None:
    cns.figure(120, 120)
    ax = cns.distplot(numeric_df, x="x", hue="group")
    assert ax is plt.gca()

    cns.figure(120, 120)
    ax2 = cns.kdeplot(numeric_df, x="x", add_mode=True)
    assert any(line.get_linestyle() == "--" for line in ax2.lines)

    cns.figure(120, 120)
    ax3 = cns.kdeplot(
        categorical_df.rename(columns={"value": "score"}), x="score", hue="hue"
    )
    assert ax3.get_legend() is not None
    assert "Anderson-Darling test" in capsys.readouterr().out

    cns.figure(120, 120)
    ax4 = cns.histplot(data=numeric_df, x="x", kde=True)
    assert ax4 is plt.gca()

    cns.figure(120, 120)
    ax5 = cns.ridgeplot(categorical_df, x="value", y="group", cmap="viridis")
    assert ax5.get_xlabel() == "value"

    cns.figure(120, 120)
    ax6 = cns.qqplot(numeric_df, x="x")
    assert ax6 is plt.gca()


def test_line_scatter_reg_and_slope_plots(
    numeric_df: pd.DataFrame,
    line_df: pd.DataFrame,
) -> None:
    cns.figure(120, 120)
    ax = cns.regplot(numeric_df, x="x", y="y")
    assert "\\rho" in ax.texts[0].get_text()

    cns.figure(120, 120)
    ax2 = cns.regplot(numeric_df, x="x", y="y", hue="group", s=5)
    assert ax2.get_legend() is not None

    cns.figure(120, 120)
    ax3 = cns.regplot(numeric_df, x="x", y="y", color="color_group")
    assert ax3.get_legend() is not None

    cns.figure(120, 120)
    ax4 = cns.scatterplot(numeric_df, x="x", y="y", hue="group", s=10)
    assert ax4.get_legend() is not None

    cns.figure(120, 120)
    ax5 = cns.lineplot(data=line_df, x="time", y="value", hue="condition")
    assert ax5.get_legend() is not None

    slope_df = pd.DataFrame(
        {
            "site": ["site1", "site1", "site2", "site2", "site3", "site3"],
            "value": [1.0, 2.0, 2.0, 1.0, 1.5, 1.7],
            "label": ["healthy", "disease"] * 3,
        }
    )
    cns.figure(120, 120)
    ax6 = cns.slopeplot(slope_df, x="site", y="value", hue="label")
    assert ax6.get_legend() is not None


def test_sets_and_specialized_plots(
    sets_fixture: dict[str, set[int]],
    sankey_df: pd.DataFrame,
    roc_df: pd.DataFrame,
) -> None:
    cns.figure(120, 120)
    axes = cns.upsetplot(sets_fixture, min_subset_size=1)
    assert set(axes) >= {"matrix", "intersections"}

    cns.figure(120, 120)
    venn = cns.vennplot(list(sets_fixture.values())[:2], labels=["A", "B"])
    assert venn is not None

    cns.figure(120, 120)
    ax = cns.sankeyplot(sankey_df, x="source", y="target")
    assert ax is plt.gca()

    cns.figure(120, 120)
    ax2 = cns.rocplot(roc_df, "truth", "model_a")
    assert len(ax2.lines) == 2

    cns.figure(120, 120)
    ax3 = cns.rocplot(roc_df, "truth", ["model_a", "model_b"])
    assert len(ax3.lines) == 3
